#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Locating modules and package source on disk.

The actual import of a module would already execute code that changes things.
Imagine a module that does ``os.system()``, it would be done during
compilation. People often connect to databases, and these kind of things, at
import time.

Therefore CPython exhibits the interfaces in an ``imp`` module in standard
library, which one can use those to know ahead of time, what file import would
load. For us unfortunately there is nothing in CPython that is easily
accessible and gives us this functionality for packages and search paths
exactly like CPython does, so we implement here a multi step search process
that is compatible.

This approach is much safer of course and there is no loss. To determine if
it's from the standard library, one can abuse the attribute ``__file__`` of the
``os`` module like it's done in ``isStandardLibraryPath`` of this module.

"""

import collections
import os
import sys
import zipfile

from nuitka.__past__ import iter_modules
from nuitka.containers.Namedtuples import makeNamedtupleClass
from nuitka.containers.OrderedSets import OrderedSet
from nuitka.Errors import NuitkaCodeDeficit
from nuitka.options.Options import (
    getMainEntryPointFilenames,
    getOutputFolderName,
    hasPythonFlagNoCurrentDirectoryInPath,
    shallExplainImports,
)
from nuitka.OutputDirectories import getSourceDirectoryPath
from nuitka.plugins.Hooks import (
    decideRecompileExtensionModules,
    getPackageExtraScanPaths,
    suppressUnknownImportWarning,
)
from nuitka.PythonFlavors import isMonolithPy
from nuitka.PythonVersions import python_version
from nuitka.SourceCodeReferences import makeSourceReferenceFromFilename
from nuitka.States import states
from nuitka.Tracing import recursion_logger
from nuitka.tree.ReformulationMultidist import locateMultidistModule
from nuitka.utils.AppDirs import getCacheDir
from nuitka.utils.FileOperations import (
    areSamePaths,
    getFileContentsHash,
    getNormalizedPath,
    getNormalizedPathJoin,
    listDir,
    removeDirectory,
)
from nuitka.utils.Importing import (
    builtin_module_names,
    getExtensionModuleSuffixes,
    getModuleFilenameSuffixes,
    getPackageDirFilename,
    isBuiltinModuleName,
)
from nuitka.utils.ModuleNames import ModuleName
from nuitka.utils.SharedLibraries import (
    hasUniversalOrMatchingMacOSArchitecture,
)
from nuitka.utils.Utils import (
    getLaunchingNuitkaProcessEnvironmentValue,
    isMacOS,
    isWin32OrPosixWindows,
)

from .IgnoreListing import isIgnoreListedNotExistingModule
from .PreloadedPackages import getPreloadedPackagePath, isPreloadedPackagePath
from .StandardLibrary import isStandardLibraryPath

# Debug traces, enabled via --explain-imports
_debug_module_finding = None

# Do not add current directory to search path.
_safe_path = None


# Debug mode
def setupImportingFromOptions():
    """Set up the importing layer from giving options."""

    # singleton, pylint: disable=global-statement
    global _debug_module_finding
    _debug_module_finding = shallExplainImports()

    global _safe_path
    _safe_path = hasPythonFlagNoCurrentDirectoryInPath()

    # Lets try and have this complete, please report failures.
    if states.is_debug and not isMonolithPy():
        _checkRaisingBuiltinComplete()

    if getOutputFolderName() is not None:
        source_dir = getSourceDirectoryPath(onefile=False, create=False)
    else:
        source_dir = None

    for filename in getMainEntryPointFilenames():
        # Inform the importing layer about the main script directory, so it can use
        # it when attempting to follow imports.

        main_dir = os.path.dirname(os.path.abspath(filename))

        if source_dir is not None and areSamePaths(source_dir, main_dir):
            continue

        addMainScriptDirectory(main_dir)


def _checkRaisingBuiltinComplete():
    for module_name in builtin_module_names:
        assert module_name in _stdlib_module_raises, module_name


warned_about = set()

# Directory where the main script lives. Should attempt to import from there.
_main_paths = OrderedSet()

# Additions to sys.paths from plugins.
_extra_paths = OrderedSet()

ModuleUsageAttempt = makeNamedtupleClass(
    "ModuleUsageAttempt",
    (
        "module_name",
        "filename",
        "module_kind",
        "finding",
        "level",
        "source_ref",
        "reason",
    ),
)


def makeModuleUsageAttempt(
    module_name, filename, module_kind, finding, level, source_ref, reason
):
    assert source_ref is not None

    # The looked for module usage attempt, cannot be a relative module name
    assert module_name.getTopLevelPackageName() != "", source_ref

    return ModuleUsageAttempt(
        module_name=module_name,
        filename=filename,
        module_kind=module_kind,
        finding=finding,
        level=level,
        source_ref=source_ref,
        reason=reason,
    )


def makeParentModuleUsagesAttempts(module_usage_attempt):
    result = []

    for parent_package_name in module_usage_attempt.module_name.getParentPackageNames():
        (
            _parent_package_name,
            parent_module_filename,
            parent_module_kind,
            parent_module_finding,
        ) = locateModule(
            module_name=parent_package_name,
            parent_package=None,
            level=0,
        )

        result.append(
            makeModuleUsageAttempt(
                module_name=parent_package_name,
                filename=parent_module_filename,
                finding=parent_module_finding,
                module_kind=parent_module_kind,
                level=0,
                source_ref=module_usage_attempt.source_ref,
                reason="import path parent",
            )
        )

    result.append(module_usage_attempt)

    return tuple(result)


def addMainScriptDirectory(main_dir):
    """Initialize the main script directory.

    We use this as part of the search path for modules.
    """
    _main_paths.add(main_dir)


def addExtraSysPaths(directories):
    for directory in directories:
        assert os.path.isdir(directory), directory

        _extra_paths.add(directory)


def getExtraSysPaths():
    return _extra_paths


def hasMainScriptDirectory():
    return bool(_main_paths)


def isPackageDir(dirname):
    """Decide if a directory is a package.

    Before Python3 it's required to have a "__init__.py" file, but then
    it became impossible to decide, and for extra fun, there is also the
    extra packages provided via "*.pth" file tricks by "site.py" loading.
    """

    return (
        "." not in os.path.basename(dirname)
        and os.path.isdir(dirname)
        and (
            python_version >= 0x300
            or isPreloadedPackagePath(dirname)
            or getPackageDirFilename(dirname) is not None
        )
    )


_package_module_name_cache = {}


def isPackageModuleName(module_name):
    """Decide if the give n module name is a package or not.

    module, Cached as it involves disk access."""

    if module_name not in _package_module_name_cache:
        _module_name, module_filename, _module_kind, _finding = locateModule(
            module_name=module_name,
            parent_package=None,
            level=0,
        )

        (
            _main_added,
            is_package,
            _is_namespace,
            _source_ref,
            _source_filename,
        ) = decideModuleSourceRef(
            filename=module_filename,
            module_name=module_name,
            is_main=False,
            is_fake=False,
            logger=None,
        )

        _package_module_name_cache[module_name] = is_package

    return _package_module_name_cache[module_name]


def getModuleNameAndKindFromFilename(module_filename):
    """Given a filename, decide the module name and kind.

    Args:
        module_name - file path of the module
    Returns:
        Tuple with the name of the module basename, and the kind of the
        module derived from the file suffix. Can be None, None if is is not a
        known file suffix.
    Notes:
        This doesn't concern itself with packages, that needs to be tracked
        by the using code. It cannot be decided from the filename at all.
    """

    if module_filename.endswith(".py"):
        return ModuleName(os.path.basename(module_filename)[:-3]), "py"

    if module_filename.endswith(".pyc"):
        return ModuleName(os.path.basename(module_filename)[:-4]), "pyc"

    # TODO: This sorting is necessary for at least Python2, but probably
    # elsewhere too, need to be sure we follow the scan order of core Python
    # here.
    for suffix in sorted(getExtensionModuleSuffixes(), key=lambda suffix: -len(suffix)):
        if module_filename.endswith(suffix):
            return (
                ModuleName(os.path.basename(module_filename)[: -len(suffix)]),
                "extension",
            )

    if os.path.isdir(module_filename):
        package_filename = getPackageDirFilename(module_filename)

        if package_filename is not None:
            for suffix in getExtensionModuleSuffixes():
                if package_filename.endswith(suffix):
                    return (
                        ModuleName(os.path.basename(module_filename)),
                        "extension",
                    )

        return ModuleName(os.path.basename(module_filename)), "py"

    return None, None


def isIgnoreListedImportMaker(source_ref):
    if isMonolithPy():
        return True

    return isStandardLibraryPath(source_ref.getFilename())


def warnAboutNotFoundImport(importing, module_name, level, source_ref):
    # This probably should not be dealt with here
    if module_name == "":
        return

    if not isIgnoreListedNotExistingModule(
        module_name
    ) and not isIgnoreListedImportMaker(source_ref):
        key = module_name, level

        if key not in warned_about:
            warned_about.add(key)

            if suppressUnknownImportWarning(
                importing=importing, source_ref=source_ref, module_name=module_name
            ):
                return

            if level == 0:
                level_desc = "as absolute import"
            elif level == -1:
                level_desc = "as relative or absolute import"
            elif level == 1:
                level_desc = "%d package level up" % level
            else:
                level_desc = "%d package levels up" % level

            if _debug_module_finding:
                if importing.getFullName().getPackageName() is not None:
                    recursion_logger.warning(
                        "%s: Cannot find '%s' in package '%s' %s."
                        % (
                            importing.getSourceReference().getAsString(),
                            module_name,
                            importing.getFullName().getPackageName().asString(),
                            level_desc,
                        )
                    )
                else:
                    recursion_logger.warning(
                        "%s: Cannot find '%s' %s."
                        % (
                            importing.getSourceReference().getAsString(),
                            module_name,
                            level_desc,
                        )
                    )


def normalizePackageName(module_name):
    # The "os.path" is strangely hacked into the "os" module, dispatching per
    # platform, we either cannot look into it, or we require that we resolve it
    # here correctly.
    if module_name == "os.path":
        module_name = ModuleName(os.path.basename(os.path.__name__))

    return module_name


# Avoid creating this over and over
_find_module_not_found = (None, None, None, None, "not-found")


def findModule(module_name, parent_package, level, logger):
    """Find a module with given package name as parent.

    The package name can be None of course. Level is the same
    as with "__import__" built-in. Warnings are optional.

    Returns:
        Returns a triple of package name the module is in, filename of
        it, which can be a directory for packages, and the location
        method used.
    """
    # We have many branches here, because there are a lot of cases to try.
    # pylint: disable=too-many-branches,too-many-return-statements

    assert type(module_name) is ModuleName, module_name

    if logger is not None:
        logger.info(
            "findModule: Enter to search %r in package %r level %s."
            % (module_name, parent_package, level)
        )

    # Do not allow star imports to get here. We just won't find modules with
    # that name, but it would be wasteful.
    assert module_name != "*"

    if level > 1:
        # TODO: Should give a warning and return not found if the levels
        # exceed the package name.
        if parent_package is not None:
            parent_package = parent_package.getRelativePackageName(level - 1)
        else:
            return _find_module_not_found

    if level == 1 and not module_name:
        # Not actually allowed, but we only catch that at run-time.
        if parent_package is None:
            return _find_module_not_found

        module_name = parent_package
        parent_package = None
        level = 0

    # Try relative imports first if we have a parent package.
    if level != 0 and parent_package is not None:
        if module_name:
            full_name = ModuleName(parent_package + "." + module_name)
        else:
            full_name = ModuleName(parent_package)

        full_name = normalizePackageName(full_name)

        preloaded_path = getPreloadedPackagePath(module_name)

        if preloaded_path is not None:
            for module_filename in preloaded_path:
                if os.path.exists(module_filename):
                    break
            else:
                module_filename = None

            return full_name, full_name.getPackageName(), module_filename, "py", "pth"

        try:
            found_module_name, module_filename, module_kind = _findModule(
                module_name=full_name,
                logger=logger,
            )
        except ImportError:
            # For relative import, that is OK, we will still try absolute.
            pass
        else:
            if logger is not None:
                logger.info(
                    "findModule: Relative imported module '%s' as '%s' in filename '%s':"
                    % (module_name, full_name, module_filename)
                )

            return (
                found_module_name,
                full_name.getPackageName(),
                module_filename,
                module_kind,
                "relative",
            )

    if level < 1 and module_name:
        module_name = normalizePackageName(module_name)

        package_name = module_name.getPackageName()

        preloaded_path = getPreloadedPackagePath(module_name)

        if preloaded_path is not None:
            for module_filename in preloaded_path:
                if os.path.exists(module_filename):
                    break
            else:
                module_filename = None

            return module_name, package_name, module_filename, "py", "pth"

        try:
            found_module_name, module_filename, module_kind = _findModule(
                module_name=module_name,
                logger=logger,
            )
        except ImportError:
            # For relative import, that is OK, we will still try absolute.
            pass
        else:
            if logger is not None:
                logger.info(
                    "findModule: Found absolute imported module '%s' in filename '%s':"
                    % (found_module_name, module_filename)
                )

            return (
                found_module_name,
                package_name,
                module_filename,
                module_kind,
                "absolute",
            )

    return _find_module_not_found


# Some platforms are case insensitive.
case_sensitive = not isMacOS() and not isWin32OrPosixWindows()

ImportScanFinding = collections.namedtuple(
    "ImportScanFinding",
    ("found_as", "found_in", "module_type", "priority", "full_path", "search_order"),
)

_list_dir_cache = {}


def listDirCached(path):
    """Cached listing of a directory."""

    if path not in _list_dir_cache:
        _list_dir_cache[path] = tuple(listDir(path))

    return _list_dir_cache[path]


def flushImportCache():
    """Clear import related caches.

    In some situations, e.g. during package rebuild, we scan and then decide to remove
    files and scan again. This allows that. Nothing in standard Nuitka should do it,
    as it throws away so much.
    """
    _list_dir_cache.clear()
    module_search_cache.clear()


def _findModuleInPath3(
    module_name,
    package_name,
    search_path,
    search_index,
    search_path_entry,
    package_directory,
):
    # First, check for a package with an init file, that would be the
    # first choice.

    # Higher values are lower priority.
    priority_map = {
        "PY_COMPILED": 3,
        "PY_SOURCE": 2,
        "C_EXTENSION": 1,
    }

    if os.path.isdir(package_directory):
        found = False

        for suffix, module_type in getModuleFilenameSuffixes():
            package_file_name = "__init__" + suffix

            file_path = getNormalizedPathJoin(package_directory, package_file_name)

            if os.path.isfile(file_path):
                yield (
                    ImportScanFinding(
                        found_as=ModuleName.makeModuleNameInPackage(
                            module_name, package_name
                        ),
                        found_in=search_path_entry,
                        module_type=module_type,
                        priority=priority_map[module_type],
                        full_path=package_directory,
                        search_order=search_index,
                    )
                )
                found = True

        if not found and python_version >= 0x300:
            yield (
                ImportScanFinding(
                    found_as=ModuleName.makeModuleNameInPackage(
                        module_name, package_name
                    ),
                    found_in=search_path_entry,
                    module_type=10,
                    priority=10,
                    full_path=package_directory,
                    search_order=search_index + len(search_path),
                )
            )

    # Then, check out suffixes of all kinds, but only for one directory.
    if search_path_entry is not None:
        last_module_type = 0
        for suffix, module_type in getModuleFilenameSuffixes():
            # Use first match per kind only.
            if module_type == last_module_type:
                continue

            full_path = getNormalizedPathJoin(search_path_entry, module_name + suffix)

            if os.path.isfile(full_path):
                yield (
                    ImportScanFinding(
                        found_as=ModuleName.makeModuleNameInPackage(
                            module_name, package_name
                        ),
                        found_in=search_path_entry,
                        module_type=module_type,
                        priority=4 + priority_map[module_type],
                        full_path=full_path,
                        search_order=search_index,
                    )
                )
                last_module_type = module_type


def _getSetuptoolsDistutilsPackageDir():
    setuptools_package_dir = locateModule(
        ModuleName("setuptools"), parent_package=None, level=0
    )[1]

    if setuptools_package_dir is not None:
        setuptools_package_dir = getNormalizedPathJoin(
            setuptools_package_dir, "_distutils"
        )

    return setuptools_package_dir


_recompile_extension_modules = {}


def getRecompileDecisionReason(module_name):
    return _recompile_extension_modules.get(module_name, (None, None))


def _findModuleInPath2(package_name, module_name, search_path, logger):
    """This is out own module finding low level implementation.

    Just the full module name and search path are given. This is then
    tasked to raise "ImportError" or return a path if it finds it, or
    None, if it is a built-in.
    """
    # We have many branches here, because there are a lot of cases to try.
    # pylint: disable=too-many-branches

    # We may have to decide between package and module, therefore build
    # a list of candidates.
    candidates = OrderedSet()

    considered = set()

    for count, search_path_entry in enumerate(search_path):
        # Don't try again, just with an entry of different casing or complete
        # duplicate.
        if os.path.normcase(search_path_entry) in considered:
            continue
        considered.add(os.path.normcase(search_path_entry))

        package_directory = getNormalizedPathJoin(
            search_path_entry, module_name.asPath()
        )

        for candidate in _findModuleInPath3(
            module_name=module_name,
            package_name=package_name,
            search_path=search_path,
            search_index=count,
            search_path_entry=search_path_entry,
            package_directory=package_directory,
        ):
            candidates.add(candidate)

    # TODO: Make this a plugin decision instead.
    if not candidates and package_name is None and module_name == "distutils":
        setuptools_package_dir = _getSetuptoolsDistutilsPackageDir()

        if setuptools_package_dir is not None:
            for candidate in _findModuleInPath3(
                module_name=module_name,
                package_name=None,
                search_path=[],
                search_path_entry=None,
                package_directory=setuptools_package_dir,
                search_index=-1,
            ):
                candidates.add(candidate)

    if logger is not None:
        logger.info("Candidates: %r" % candidates)

    found_candidate = None

    if candidates:
        # Sort by priority, with entries from same path element coming first, then desired type. In case
        # pf there being an extension module,
        if any(c.module_type == "C_EXTENSION" for c in candidates):

            def prioritize(c):
                if c.module_type == "PY_SOURCE":
                    decision, reason = decideRecompileExtensionModules(c.found_as)
                    _recompile_extension_modules[c.found_as] = decision, reason
                else:
                    decision = False

                return (
                    c.search_order,
                    ((c.priority - 2) if decision else c.priority),
                )

        else:

            def prioritize(c):
                return (c.search_order, c.priority)

        candidates = tuple(sorted(candidates, key=prioritize))

        # On case sensitive systems, no resolution needed.
        if case_sensitive:
            found_candidate = candidates[0]
        else:
            for candidate in candidates:
                if candidate.found_in is None:
                    found_candidate = candidate
                else:
                    for fullname, _filename in listDirCached(candidate.found_in):
                        if fullname == candidate.full_path:
                            found_candidate = candidate
                            break

                if found_candidate:
                    break

            # Only exact case matches matter, all candidates were ignored,
            # lets just fall through to raising the import error.

    if found_candidate is None:
        # Nothing found.
        raise ImportError
    if (
        found_candidate.module_type == "C_EXTENSION"
        and isMacOS()
        and not hasUniversalOrMatchingMacOSArchitecture(
            getPackageDirFilename(found_candidate.full_path)
            if os.path.isdir(found_candidate.full_path)
            else found_candidate.full_path
        )
    ):
        # Not usable for target architecture.
        raise ImportError

    return (
        found_candidate.found_as,
        found_candidate.full_path,
        "extension" if found_candidate.module_type == "C_EXTENSION" else "py",
    )


_egg_files = {}


def _unpackPathElement(path_entry):
    if not path_entry:
        return "."  # empty means current directory

    if os.path.isfile(path_entry):
        if path_entry.lower().endswith((".egg", ".zip")):
            if path_entry not in _egg_files:
                checksum = getFileContentsHash(path_entry)

                target_dir = getNormalizedPathJoin(getCacheDir("egg-content"), checksum)

                if not os.path.exists(target_dir):
                    try:
                        # Not all Python versions allow using with here, pylint: disable=consider-using-with
                        zip_ref = zipfile.ZipFile(path_entry, "r")
                        zip_ref.extractall(target_dir)
                        zip_ref.close()
                    except BaseException:
                        removeDirectory(
                            target_dir,
                            logger=recursion_logger,
                            ignore_errors=True,
                            extra_recommendation=None,
                        )
                        raise

                _egg_files[path_entry] = target_dir

            return _egg_files[path_entry]

    return path_entry


def getPythonUnpackedSearchPath():
    """Python search path with with eggs unpacked."""

    # TODO: Maybe cache this for a given "sys.path" as we do IO checks each time.
    return [_unpackPathElement(path_element) for path_element in sys.path]


def getPackageSearchPath(package_name):
    # TODO: Some branches here are hard coded things that ought to be plugin
    # decisions, pylint: disable=too-many-branches

    if not _main_paths:
        return None

    if package_name is None:
        result = []

        if not _safe_path:
            result.append(os.getcwd())

        result += list(_main_paths) + getPythonUnpackedSearchPath() + list(_extra_paths)
    elif "." in package_name:
        parent_package_name, child_package_name = package_name.splitModuleBasename()

        result = []
        for element in getPackageSearchPath(parent_package_name):
            package_dir = getNormalizedPathJoin(element, child_package_name.asPath())

            if isPackageDir(package_dir):
                result.append(package_dir)
                # Hack for "uniconvertor". TODO: Move this to plug-in decision. This
                # fails the above test, but at run time should be a package.
                # spell-checker: ignore uniconvertor
            elif package_name == "uniconvertor.app.modules":
                result.append(package_dir)

        return result

    else:
        preloaded_path = getPreloadedPackagePath(package_name)

        if preloaded_path is not None:
            return preloaded_path

        def getPackageDirCandidates(element):
            yield getNormalizedPathJoin(element, package_name.asPath()), False

            for extra_path in getPackageExtraScanPaths(package_name, element):
                yield extra_path, True

        result = []

        # TODO: Make this a plugin decision instead.
        if package_name == "distutils":
            setuptools_package_dir = _getSetuptoolsDistutilsPackageDir()

            if setuptools_package_dir is not None:
                result.append(setuptools_package_dir)

        for element in getPackageSearchPath(None):
            for package_dir, force_package in getPackageDirCandidates(element):
                if force_package or isPackageDir(package_dir):
                    result.append(package_dir)

    return OrderedSet(
        element for element in OrderedSet(result) if os.path.exists(element)
    )


def _findModuleInPath(module_name, logger):
    package_name, module_name = module_name.splitModuleBasename()

    if logger is not None:
        logger.info(
            "_findModuleInPath: Enter for %s in %s" % (module_name, package_name)
        )

    # The "site" module must be located based on PYTHONPATH before it was
    # executed, while we normally search in PYTHONPATH after it was executed,
    # and on some systems, that fails.
    if package_name is None and module_name == "site":
        candidate = getLaunchingNuitkaProcessEnvironmentValue("NUITKA_SITE_FILENAME")

        if candidate:
            return module_name, candidate, "py"

    # Free pass for built-in modules, they need not exist.
    if package_name is None and isBuiltinModuleName(module_name):
        return module_name, None, "built-in"

    # These are existing in the standard library but are effectively inline
    # already and should be avoided to look at.
    if package_name == "importlib" and module_name in (
        "_bootstrap",
        "_bootstrap_external",
    ):
        candidate = package_name.getChildNamed(module_name)
        if candidate.asString() in sys.modules:
            return candidate, None, "built-in"

    search_path = getPackageSearchPath(package_name)

    if logger is not None:
        logger.info(
            "_findModuleInPath: Using search path %s for %s"
            % (search_path, package_name)
        )

    try:
        found_module_name, module_filename, module_kind = _findModuleInPath2(
            package_name=package_name,
            module_name=module_name,
            search_path=search_path,
            logger=logger,
        )
    except SyntaxError:
        # Warn user, as this is kind of unusual.
        recursion_logger.warning(
            "%s: Module cannot be imported due to syntax errors.",
            module_name if package_name is None else package_name + "." + module_name,
        )

        return None, None, None

    if logger is not None:
        logger.info(
            "_findModuleInPath: _findModuleInPath2 gave %s %s"
            % (module_filename, module_kind)
        )

    return found_module_name, module_filename, module_kind


module_search_cache = {}


def _findModule(module_name, logger):
    # Not a good module name. TODO: Push this to ModuleName() creation maybe.
    assert module_name != ""

    if logger is not None:
        logger.info("_findModule: Enter to search '%s'." % module_name)

    assert module_name.getBasename(), module_name

    key = module_name

    if key in module_search_cache:
        result = module_search_cache[key]

        if logger is not None:
            logger.info("_findModule: Cached result (see previous call).")

        if result is ImportError:
            raise ImportError

        return result

    try:
        module_search_cache[key] = _findModuleInPath(
            module_name=module_name,
            logger=logger,
        )
    except ImportError:
        module_search_cache[key] = ImportError
        raise

    # assert len(module_search_cache[key]) == 2, (module_name, module_search_cache[key])

    return module_search_cache[key]


def locateModule(module_name, parent_package, level, logger=None):
    """Locate a module with given package name as parent.

    The package name can be None of course. Level is the same
    as with "__import__" built-in.

    Returns:
        Returns a tuple of module name the module has considering
        package containing it, and filename of it which can be a
        directory for packages, the module kind, and the finding
        kind.
    """

    if not _main_paths:
        raise NuitkaCodeDeficit(
            "Error, cannot locate modules before import mechanism is setup."
        )

    if module_name.isMultidistModuleName():
        return locateMultidistModule(module_name)

    if _debug_module_finding and logger is None:
        logger = recursion_logger

    found_module_name, module_package, module_filename, module_kind, finding = (
        findModule(
            module_name=module_name,
            parent_package=parent_package,
            level=level,
            logger=logger,
        )
    )

    # Allowing ourselves to be lazy.
    assert module_kind is not None or module_filename is None, module_name

    assert module_package is None or (
        type(module_package) is ModuleName and module_package != ""
    ), ("Must not attempt to locate %r" % module_name)

    if module_filename is not None:
        module_filename = getNormalizedPath(module_filename)

        if found_module_name == "distutils":
            _module_name, module_kind = getModuleNameAndKindFromFilename(
                module_filename
            )
            module_name = found_module_name
        else:
            module_name, module_kind = getModuleNameAndKindFromFilename(module_filename)
            module_name = ModuleName.makeModuleNameInPackage(
                module_name, module_package
            )

            if states.is_debug:
                assert module_name == found_module_name, (
                    module_name,
                    found_module_name,
                    parent_package,
                )

    elif finding == "not-found":
        if parent_package is not None:
            if not module_name:
                module_name = parent_package
            elif level != 0:
                module_name = ModuleName.makeModuleNameInPackage(
                    package_name=parent_package, module_name=module_name
                )
        elif level > 0:
            module_name = ModuleName("")

    return module_name, module_filename, module_kind, finding


def locateModules(package_name):
    """Determine child module names.

    Return:
        generator of ModuleName objects
    """
    package_name = ModuleName(package_name)

    module_filename = locateModule(
        module_name=ModuleName(package_name), parent_package=None, level=0
    )[1]

    if module_filename is not None:
        for sub_module in iter_modules([module_filename]):
            yield package_name.getChildNamed(sub_module.name)


def hasModule(module_name):
    """Tell if a module exists in the installation."""

    _module_name, _module_filename, _module_kind, finding = locateModule(
        module_name=ModuleName(module_name), parent_package=None, level=0
    )

    return finding != "not-found"


def decideModuleSourceRef(filename, module_name, is_main, is_fake, logger):
    # Many branches due to the many cases

    assert type(module_name) is ModuleName, module_name
    assert filename is not None

    is_namespace = False
    is_package = False

    if is_main and os.path.isdir(filename):
        source_filename = getNormalizedPathJoin(filename, "__main__.py")

        if not os.path.isfile(source_filename):
            sys.stderr.write(
                "%s: can't find '__main__' module in '%s'\n"
                % (os.path.basename(sys.argv[0]), filename)
            )
            sys.exit(2)

        filename = source_filename

        main_added = True
    else:
        main_added = False

    if is_fake:
        source_filename = filename

        source_ref = makeSourceReferenceFromFilename(filename=filename)

        module_name = is_fake

    elif os.path.isfile(filename):
        source_filename = filename

        source_ref = makeSourceReferenceFromFilename(filename=filename)

    elif isPackageDir(filename):
        is_package = True

        source_filename = getNormalizedPathJoin(filename, "__init__.py")

        if not os.path.isfile(source_filename):
            source_ref = makeSourceReferenceFromFilename(filename=filename).atInternal()
            is_namespace = True
        else:
            source_ref = makeSourceReferenceFromFilename(
                filename=getNormalizedPath(os.path.abspath(source_filename))
            )

    else:
        return logger.sysexit(
            "%s: can't open file '%s'." % (os.path.basename(sys.argv[0]), filename),
            exit_code=2,
        )

    return (
        main_added,
        is_package,
        is_namespace,
        source_ref,
        source_filename,
    )


# spell-checker: ignore _posixsubprocess,pyexpat,xxsubtype,_bytesio,_interpchannels
# spell-checker: ignore _interpqueues,_lsprof,_multibytecodec,_posixshmem _winapi
# spell-checker: ignore  _testbuffer _testexternalinspection _testimportmultiple
# spell-checker: ignore _testinternalcapi _testmultiphase _testsinglephase _scproxy
# spell-checker: ignore _xxtestfuzz _xxsubinterpreters imageop _xxinterpchannels
_stdlib_module_raises = {
    "_abc": False,
    "__builtin__": False,
    "_ast": False,
    "_asyncio": False,
    "_bisect": False,
    "_blake2": False,
    "_bytesio": False,
    "_bz2": False,
    "_codecs_cn": False,
    "_codecs_hk": False,
    "_codecs_iso2022": False,
    "_codecs_jp": False,
    "_codecs_kr": False,
    "_codecs_tw": False,
    "_codecs": False,
    "_collections": False,
    "_contextvars": False,
    "_crypt": False,
    "_csv": False,
    "_ctypes": False,
    "_ctypes_test": False,
    "_curses": False,
    "_curses_panel": False,
    "_datetime": False,
    "_dbm": False,
    "_decimal": False,
    "_elementtree": False,
    "_fileio": False,
    "_functools": False,
    "_hashlib": False,
    "_heapq": False,
    "_hmac": False,
    "_hotshot": False,
    "_imp": False,
    "_interpchannels": False,
    "_interpqueues": False,
    "_interpreters": False,
    "_io": False,
    "_json": False,
    "_locale": False,
    "_lsprof": False,
    "_lzma": False,
    "_md5": False,
    "_multiprocessing": False,
    "_multibytecodec": False,
    "_opcode": False,
    "_operator": False,
    "_peg_parser": False,
    "_pickle": False,
    "_posixshmem": False,
    "_posixsubprocess": False,
    "_queue": False,
    "_random": False,
    "_scproxy": False,
    "_sha": False,  # TODO: Not entirely clear if that's true
    "_sha1": False,
    "_sha2": False,
    "_sha256": False,
    "_sha3": False,
    "_sha512": False,
    "_signal": False,
    "_socket": False,
    "_sqlite3": False,
    "_sre": False,
    "_stat": False,
    "_statistics": False,
    "_ssl": True,
    "_string": False,
    "_struct": False,
    "_subprocess": False,
    "_suggestions": False,
    "_symtable": False,
    "_sysconfig": False,
    "_testbuffer": False,
    "_testexternalinspection": False,
    "_testimportmultiple": False,
    "_testinternalcapi": False,
    "_testmultiphase": False,
    "_testsinglephase": False,
    "_thread": False,
    "_tkinter": True,
    "_tokenize": False,
    "_tracemalloc": False,
    "_types": False,
    "_typing": False,
    "_uuid": False,
    "_warnings": False,
    "_weakref": False,
    "_winapi": False,
    "_winreg": False,
    "_xxtestfuzz": False,
    "_xxsubinterpreters": False,
    "_xxinterpchannels": False,
    "_zoneinfo": False,
    "array": False,
    "atexit": False,
    "audioop": False,
    "binascii": False,
    "builtins": False,
    "cPickle": False,
    "cStringIO": False,
    "cmath": False,
    "datetime": False,
    "errno": False,
    "exceptions": False,
    "faulthandler": False,
    "fcntl": False,
    "future_builtins": False,
    "gc": False,
    "grp": False,
    "imageop": False,
    "imp": False,
    "itertools": False,
    "marshal": False,
    "math": False,
    "mmap": False,
    "msvcrt": False,
    "nis": False,
    "nt": False,
    "operator": False,
    "ossaudiodev": False,
    "parser": False,
    "posix": False,
    "pwd": False,
    "pyexpat": False,
    "readline": True,
    "resource": False,
    "select": False,
    "signal": False,
    "spwd": False,
    "strop": False,
    "sys": False,
    "syslog": False,
    "termios": False,
    "thread": False,
    "time": False,
    "unicodedata": False,
    "winreg": False,
    "xxsubtype": False,
    "zipimport": False,
    "zlib": False,
}


def isNonRaisingBuiltinModule(module_name):
    assert isBuiltinModuleName(module_name), module_name

    # Return None, if we don't know.
    return _stdlib_module_raises.get(module_name)


def _getChildPackageNames(module_name):
    module_name = ModuleName(module_name)

    _module_name, module_filename, _module_kind, _finding = locateModule(
        parent_package=None, module_name=module_name, level=0
    )

    package_dir = module_filename

    for module_info in iter_modules([package_dir]):
        child_name = module_name.getChildNamed(module_info.name)
        yield child_name

        if module_info.ispkg:  # spell-checker: ignore ispkg
            for sub_module_name in getChildPackageNames(child_name):
                yield sub_module_name


def getChildPackageNames(module_name):
    return tuple(_getChildPackageNames(module_name))


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Launcher for running a script inside a container.

Podman and Docker should both work, but the first one is recommended.
"""

import os
import shutil
import sys

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.tools.release.Release import (
    getBranchName,
    getBranchRemoteIdentifier,
)
from nuitka.Tracing import OurLogger
from nuitka.utils.Download import getCachedDownloadedMinGW64
from nuitka.utils.Execution import (
    callProcess,
    check_output,
    getExecutablePath,
    withEnvironmentPathAdded,
)
from nuitka.utils.FileOperations import (
    changeFilenameExtension,
    getFileContents,
    makePath,
    putTextFileContents,
    withTemporaryDirectory,
)
from nuitka.utils.Utils import getArchitecture, isWin32Windows

from .Podman import getPodmanExecutablePath

containers_logger = OurLogger("Nuitka-Containers", base_style="blue")


def parseOptions():
    parser = makeOptionsParser(usage=None, epilog=None)

    parser.add_option(
        "--container-id",
        action="store",
        dest="container_id",
        default="CI",
        help="""
Name of the container to use. Defaults to "CI" which is used for testing
Nuitka with Linux.
""",
    )

    parser.add_option(
        "--no-build-container",
        action="store_true",
        dest="no_build_container",
        help="""
Do not update the the container, use it if updating was done recently.
""",
    )

    parser.add_option(
        "--command",
        action="store",
        dest="command",
        help="""
Command to execute, all in one value.
""",
    )

    parser.add_option(
        "--podman-path",
        action="store",
        dest="podman_path",
        default=None,
        help="""
Podman binary in case you do not have it in your path.
""",
    )

    parser.add_option(
        "--podman-verbose",
        action="store_false",
        dest="quiet",
        default=True,
        help="""
Dot not use podman quietly, giving more messages during build.""",
    )

    parser.add_option(
        "--shared-path",
        action="append",
        dest="shared_paths",
        default=[],
        help="""
Path to share with container, use "--shared-path=src=dst" format for directory names.
""",
    )

    parser.add_option(
        "--network",
        action="store_true",
        dest="network",
        default=False,
        help="""
This container run should be allowed to use network.
""",
    )

    parser.add_option(
        "--isolated",
        action="store_true",
        dest="isolated",
        default=None,
        help="""
This container run should not be provided host access of any kind.
""",
    )

    parser.add_option(
        "--no-isolated",
        action="store_false",
        dest="isolated",
        default=None,
        help="""
This container run should be provided host access even if the name suggests otherwise.
""",
    )

    parser.add_option(
        "--pbuilder",
        action="store_true",
        dest="pbuilder",
        default=False,
        help="""
This container run should be allowed to use pbuilder.
""",
    )

    options, positional_args = parser.parse_args()

    if positional_args:
        containers_logger.sysexit(
            "This command takes no positional arguments, check help output."
        )

    if options.podman_path is None:
        options.podman_path = getPodmanExecutablePath(containers_logger)

        assert options.podman_path is not None

    return options


def isPodman(podman_path):
    return "podman" in os.path.normcase(os.path.basename(podman_path))


def updateContainer(podman_path, container_tag_name, container_file_path, quiet):
    requirements_file = os.path.join(
        os.path.dirname(__file__), "..", "..", "..", "requirements-devel.txt"
    )

    if not os.path.exists(requirements_file):
        containers_logger.sysexit(
            "Error, cannot find expected requirements-devel.txt file."
        )

    requirements_private_file = os.path.join(
        os.path.dirname(requirements_file),
        "nuitka",
        "utils",
        "requirements-private.txt",
    )

    containers_logger.info("Updating container '%s'..." % container_tag_name)

    # Use a temporary directory for the build context.
    with withTemporaryDirectory(containers_logger) as build_context_dir:
        # Copy requirements-devel.txt
        shutil.copy(
            requirements_file, os.path.join(build_context_dir, "requirements-devel.txt")
        )

        # Copy nuitka/utils/requirements-private.txt
        requirements_private_dest = os.path.join(build_context_dir, "nuitka", "utils")
        makePath(requirements_private_dest)
        shutil.copy(
            requirements_private_file,
            os.path.join(requirements_private_dest, "requirements-private.txt"),
        )

        # Read original container file
        container_content = getFileContents(container_file_path)

        # Append new instructions
        container_content += "\n# Automatic requirements caching\n"
        container_content += "COPY requirements-devel.txt /etc/requirements-devel.txt\n"
        container_content += "COPY nuitka/utils/requirements-private.txt /etc/nuitka/utils/requirements-private.txt\n"

        # Install requirements
        # We try both python3 and python2.
        container_content += (
            "RUN if command -v python3; "
            "then python3 -m pip install ${NUITKA_PIP_FLAGS} -r /etc/requirements-devel.txt; "
            "fi\n"
        )
        container_content += (
            "RUN if command -v python2; "
            "then wget https://bootstrap.pypa.io/pip/2.7/get-pip.py -O /var/tmp/get-pip.py && "
            "python2 /var/tmp/get-pip.py && "
            "python2 -m pip install ${NUITKA_PIP_FLAGS} -r /etc/requirements-devel.txt; "
            "fi\n"
        )

        # Write modified content to temporary containerfile in context
        temp_container_file = os.path.join(build_context_dir, "Containerfile")
        putTextFileContents(temp_container_file, container_content)

        command = [
            podman_path,
            "build",
            # Tolerate errors checking for image download, and use old one
            "--tag",
            container_tag_name,
            "-f",
            temp_container_file,
            "--network=host",
        ]

        if quiet:
            command.append("--quiet")

        if isPodman(podman_path):
            # Podman only.
            command.append("--pull=newer")

        # Always append context
        command.append(build_context_dir)

        exit_code = callProcess(command)

        if exit_code:
            containers_logger.sysexit(
                "Failed to update container with exit code '%d'. Command used was: %s"
                % (exit_code, " ".join(command)),
                exit_code=exit_code,
            )

        containers_logger.info(
            "Updated container '%s' successfully." % container_tag_name
        )


def getCppPath():
    cpp_path = getExecutablePath("cpp_path")

    # Windows extra ball, attempt the downloaded one.
    if isWin32Windows() and cpp_path is None:
        from nuitka.options.Options import assumeYesForDownloads

        mingw64_gcc_path = getCachedDownloadedMinGW64(
            target_arch=getArchitecture(),
            assume_yes_for_downloads=assumeYesForDownloads(),
            download_ok=True,
            experimental=False,
        )

        with withEnvironmentPathAdded("PATH", os.path.dirname(mingw64_gcc_path)):
            cpp_path = getExecutablePath("cpp")

            os.environ["CPP_PATH"] = cpp_path

    if cpp_path is None:
        containers_logger.sysexit(
            "Error, need 'cpp' binary to execute this container file using.'"
        )

    return cpp_path


def _makeMountDesc(options, src_path, dst_path, flags):
    src_path = os.path.expanduser(src_path)

    mount_desc = "type=bind,source=%s,dst=%s" % (src_path, dst_path)

    if isPodman(options.podman_path):
        mount_desc += ",relabel=shared"

    if flags:
        mount_desc += ",%s" % flags

    if options.isolated:
        mount_desc += ",ro"

    return mount_desc


def _checkIsolated(options, container_tag_name):
    if options.isolated:
        containers_logger.info("Running isolated as per user choice.")
    # Auto-isolate by container name.
    elif "isolated" in container_tag_name.lower() and options.isolated is None:
        containers_logger.info("Running isolated as per container name default.")
        options.isolated = True
    elif options.isolated is False:
        containers_logger.info("Running NOT isolated as per user choice.")
    else:
        containers_logger.info("Running NOT isolated as per default.")


def _checkContainerArgument(options, default_container_directory):
    if ("/" in options.container_id or "\\" in options.container_id) and os.path.exists(
        options.container_id
    ):
        container_file_path = options.container_id

        if container_file_path.endswith(".in"):
            container_file_path_template = container_file_path
            container_file_path = container_file_path[:-3]
        else:
            container_file_path_template = None

        options.container_id = changeFilenameExtension(
            os.path.basename(container_file_path), ""
        )
    else:
        container_file_path = os.path.join(
            default_container_directory, options.container_id + ".containerfile"
        )
        container_file_path_template = container_file_path + ".in"

    return container_file_path_template, container_file_path


def main():
    options = parseOptions()

    if options.command is None:
        options.command = "python3 -m nuitka --version"

    containers_logger.info(
        "Running in container '%s' this command: %s"
        % (options.container_id, options.command)
    )

    default_container_directory = os.path.join(os.path.dirname(__file__), "containers")

    container_file_path_template, container_file_path = _checkContainerArgument(
        options=options, default_container_directory=default_container_directory
    )

    if container_file_path_template is not None and os.path.isfile(
        container_file_path_template
    ):
        # Check requirement.
        cpp_path = getCppPath()
        command = [
            cpp_path,
            "-E",
            "-I",
            default_container_directory,
            container_file_path_template,
        ]

        output = check_output(command, shell=False)
        if str is not bytes:
            output = output.decode("utf8")

        putTextFileContents(container_file_path, output, encoding="utf8")

    if not os.path.isfile(container_file_path):
        containers_logger.sysexit(
            "Error, no container ID '%s' found at '%s'."
            % (options.container_id, container_file_path)
        )

    getBranchRemoteIdentifier()

    container_tag_name = "nuitka-build-%s-%s:latest" % (
        options.container_id.lower(),
        getBranchRemoteIdentifier() + "-" + getBranchName(),
    )

    _checkIsolated(options, container_tag_name)

    if not options.no_build_container:
        updateContainer(
            podman_path=options.podman_path,
            container_tag_name=container_tag_name,
            container_file_path=container_file_path,
            quiet=options.quiet,
        )

    command = [options.podman_path, "run", "--rm"]

    command.extend(
        (
            "--mount",
            _makeMountDesc(options=options, src_path=".", dst_path="/src", flags=""),
        )
    )

    if options.network:
        command.append("--add-host=ssh.nuitka.net:116.202.30.188")
    else:
        command.append("--network=none")

    # May need to allow pbuilder to create device nodes, makes the container insecure
    # though.
    if options.pbuilder:
        command += ["--privileged"]

    for path_desc in options.shared_paths:
        if path_desc.count("=") == 1:
            src_path, dst_path = path_desc.split("=")
            flags = ""
        else:
            src_path, dst_path, flags = path_desc.split("=", 2)

        src_path = os.path.expanduser(src_path)

        command.extend(
            (
                "--mount",
                _makeMountDesc(
                    options=options, src_path=src_path, dst_path=dst_path, flags=flags
                ),
            )
        )

    # Interactive if possible only.
    if sys.stdout.isatty():
        command.append("-it")

    command += [
        container_tag_name,
        "bash",
        "-l",
        "-c",
        "cd /src;" + options.command + "; exit $?",
    ]

    exit_code = callProcess(command, shell=False, logger=containers_logger)

    containers_logger.sysexit(
        "Finished container run with exit code '%d'." % exit_code, exit_code=exit_code
    )


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""C labels, small helpers.

Much things are handled with "goto" statements in the generated code, error
exits, finally blocks, etc. this provides just the means to emit a label or
the goto statement itself.
"""

from nuitka.utils.CStrings import encodePythonStringToC


def getGotoCode(label, emit):
    assert label is not None

    emit("goto %s;" % label)


def getLabelCode(label, emit):
    assert label is not None

    emit("%s:;" % label)


def getBranchingCode(condition, emit, context):
    true_target = context.getTrueBranchTarget()
    false_target = context.getFalseBranchTarget()

    if true_target is not None and false_target is None:
        emit("if (%s) goto %s;" % (condition, true_target))
    elif true_target is None and false_target is not None:
        emit("if (!(%s)) goto %s;" % (condition, false_target))
    else:
        assert true_target is not None and false_target is not None

        emit(
            """\
if (%s) {
    goto %s;
} else {
    goto %s;
}"""
            % (condition, true_target, false_target)
        )


def getStatementTrace(source_desc, statement_repr):
    # return r'PRINT_STRING("Execute: " %s "\n");' % (
    return 'NUITKA_PRINT_TRACE("Execute: " %s);' % (
        encodePythonStringToC(source_desc + b" " + statement_repr),
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Code generation for type alias statement helpers."""

from .CodeHelpers import (
    generateChildExpressionCode,
    withObjectCodeTemporaryAssignment,
)
from .ErrorCodes import getErrorExitCode
from .TupleCodes import getTupleCreationCode


def generateTypeAliasCode(to_name, expression, emit, context):
    type_params_name = context.allocateTempName("type_params_value")
    getTupleCreationCode(
        to_name=type_params_name,
        elements=expression.subnode_type_params,
        emit=emit,
        context=context,
    )

    compute_value_name = generateChildExpressionCode(
        expression=expression.subnode_value, emit=emit, context=context
    )

    type_alias_name = expression.subnode_name.getVariableName()

    with withObjectCodeTemporaryAssignment(
        to_name, "type_alias_value", expression, emit, context
    ) as value_name:
        emit(
            "%s = MAKE_TYPE_ALIAS(%s, %s, %s, %s);"
            % (
                value_name,
                context.getConstantCode(constant=type_alias_name),
                type_params_name,
                compute_value_name,
                context.getConstantCode(constant=context.getModuleName().asString()),
            )
        )

        getErrorExitCode(
            check_name=value_name,
            release_names=(type_alias_name, compute_value_name, type_params_name),
            emit=emit,
            context=context,
            needs_check=expression.mayRaiseExceptionOperation(),
        )

        context.addCleanupTempName(value_name)


def generateTypeVarCode(to_name, expression, emit, context):
    with withObjectCodeTemporaryAssignment(
        to_name, "type_var_value", expression, emit, context
    ) as value_name:
        if expression.isExpressionTypeVariable():
            # TypeVar
            helper_function = "MAKE_TYPE_VAR"
        elif expression.isExpressionTypeVariableTuple():
            # TypeVarTuple
            helper_function = "MAKE_TYPE_VAR_TUPLE"
        else:
            # ParamSpec
            assert expression.isExpressionParameterSpecification()
            helper_function = "MAKE_PARAM_SPEC"

        emit(
            "%s = %s(tstate, %s);"
            % (
                value_name,
                helper_function,
                context.getConstantCode(constant=expression.name),
            )
        )

        getErrorExitCode(
            check_name=value_name,
            emit=emit,
            context=context,
            needs_check=False,
        )

        context.addCleanupTempName(value_name)


def generateTypeGenericCode(to_name, expression, emit, context):
    type_params_value = generateChildExpressionCode(
        expression=expression.subnode_type_params, emit=emit, context=context
    )

    with withObjectCodeTemporaryAssignment(
        to_name, "type_params_value", expression, emit, context
    ) as value_name:
        emit(
            "%s = MAKE_TYPE_GENERIC(tstate, %s);"
            % (
                value_name,
                type_params_value,
            )
        )

        getErrorExitCode(
            check_name=value_name,
            release_name=type_params_value,
            emit=emit,
            context=context,
        )

        context.addCleanupTempName(value_name)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Automatic formatting of JSON files."""

import os

from nuitka.utils.Download import getCachedDownload
from nuitka.utils.Execution import check_call, getNullOutput
from nuitka.utils.FileOperations import addFileExecutablePermission
from nuitka.utils.Utils import getArchitecture, getOS

_biome_path = None


def _getBiomeBinaryPath(assume_yes_for_downloads):
    """Downloads and returns the path to the biome executable.

    Args:
        assume_yes_for_downloads: bool - if tools should be downloaded automatically

    Returns:
        str or None: path to the biome executable or None if not supported
    """
    # Many branches used to map OS and arch, pylint: disable=too-many-branches

    # singleton, pylint: disable=global-statement
    global _biome_path

    if _biome_path is not None:
        return _biome_path

    os_name = getOS()
    architecture = getArchitecture()

    if os_name == "Linux":
        if architecture == "x86_64":
            arch_name = "x64"
        elif architecture == "arm64":
            arch_name = "arm64"
        else:
            return None
        binary_name = "biome-linux-%s" % arch_name
    elif os_name == "Windows":
        if architecture == "x86_64":
            arch_name = "x64"
        elif architecture == "x86":
            arch_name = "ia32"
        elif architecture == "arm64":
            arch_name = "arm64"
        else:
            return None
        binary_name = "biome-win32-%s.exe" % arch_name
    elif os_name == "Darwin":
        if architecture == "x86_64":
            arch_name = "x64"
        elif architecture == "arm64":
            arch_name = "arm64"
        else:
            return None
        binary_name = "biome-darwin-%s" % arch_name
    else:
        return None

    version = "2.2.4"
    url = (
        "https://github.com/biomejs/biome/releases/download/%%40biomejs%%2Fbiome%%40%s/%s"
        % (
            version,
            binary_name,
        )
    )

    _biome_path = getCachedDownload(
        name="biome",
        url=url,
        is_arch_specific=getArchitecture(),
        specificity=version,
        binary=os.path.basename(url),
        unzip=False,
        flatten=False,
        message="Nuitka will make use of 'biome' to format JSON files.",
        reject=None,
        assume_yes_for_downloads=assume_yes_for_downloads,
        download_ok=True,
    )

    addFileExecutablePermission(_biome_path)

    return _biome_path


def formatJsonFile(filename, assume_yes_for_downloads):
    """Format a JSON file using the biome formatter.

    Args:
        filename: str - path to the JSON file
        assume_yes_for_downloads: bool - if tools should be downloaded automatically
    """
    # Many tools work on files, and when they do, they need to be told to
    # treat it as a JSON file.
    biome_path = _getBiomeBinaryPath(assume_yes_for_downloads=assume_yes_for_downloads)

    if biome_path:
        command = (
            biome_path,
            "format",
            "--write",
            "--json-formatter-expand=always",
            "--json-formatter-indent-style=space",
            "--log-level=warn",
            "--config-path=%s" % os.path.join(os.path.dirname(__file__), "biome.json"),
            "--",
            filename,
        )

        with getNullOutput() as null_output:
            check_call(command, stdout=null_output)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""
This module handles the extraction of build configuration from 'build' backend projects.
"""

import os
import sys

from nuitka.importing.Importing import addMainScriptDirectory
from nuitka.PythonVersions import getSitePackageCandidateNames
from nuitka.utils.Execution import (
    executeProcess,
    withEnvironmentVarsOverridden,
)
from nuitka.utils.FileOperations import (
    makePath,
    putTextFileContents,
    withDirectoryChange,
    withTemporaryDirectory,
)
from nuitka.utils.Json import loadJsonFromFilename

from .BuildPackageCommon import reportBuildError, setProjectName


def getBuildBackendConfiguration(logger):
    """
    Get the build configuration from a project using the 'build' backend.
    """
    # This is "build" handling.
    with withTemporaryDirectory(
        prefix="nuitka-project-dump", logger=logger
    ) as temp_dir:

        mock_site_packages_path = os.path.join(temp_dir, "site_packages")
        makePath(mock_site_packages_path)

        nuitka_path = os.path.dirname(os.path.dirname(sys.modules["nuitka"].__file__))

        # spell-checker: ignore sitecustomize,cmdclass
        putTextFileContents(
            filename=os.path.join(mock_site_packages_path, "sitecustomize.py"),
            contents="""
import sys
import os

try:
    import wheel.bdist_wheel
except ImportError:
    sys.stderr.write("Wheel not importable in sitecustomize\\n")
    os._exit(1)

# Ensure nuitka is importable
nuitka_path = %(nuitka_path)r
if nuitka_path not in sys.path:
    sys.path.insert(0, nuitka_path)
    added_path = True
else:
    added_path = False

try:
    from nuitka.distutils.DistutilsCommands import build as nuitka_build
except ImportError:
    sys.stderr.write("Nuitka not importable in sitecustomize\\n")
    os._exit(1)

if added_path:
    del sys.path[0]


import setuptools
import distutils.core
import setuptools.command.egg_info

# Patch setuptools
_old_setuptools_setup = setuptools.setup
def new_setuptools_setup(**attrs):
    attrs['cmdclass'] = attrs.get('cmdclass', {})
    attrs['cmdclass']['build'] = nuitka_build
    return _old_setuptools_setup(**attrs)
setuptools.setup = new_setuptools_setup

# Patch distutils
_old_distutils_setup = distutils.core.setup
def new_distutils_setup(**attrs):
    attrs['cmdclass'] = attrs.get('cmdclass', {})
    attrs['cmdclass']['build'] = nuitka_build
    return _old_distutils_setup(**attrs)
distutils.core.setup = new_distutils_setup

# Force egg_base to temp_dir
_old_egg_info_initialize_options = setuptools.command.egg_info.egg_info.initialize_options
def new_egg_info_initialize_options(self):
    _old_egg_info_initialize_options(self)
    self.egg_base = %(egg_base)r
setuptools.command.egg_info.egg_info.initialize_options = new_egg_info_initialize_options
"""
            % {
                "nuitka_path": nuitka_path,
                "egg_base": temp_dir,
            },
        )

        dump_filename = os.path.join(temp_dir, "build_config.json")

        # We use the temporary directory for the project to avoid implicit imports
        # of the current directory, which may contain a "build" folder that clashes.
        project_dir = os.getcwd()

        with withEnvironmentVarsOverridden(
            {
                "NUITKA_DUMP_BUILD_CONFIG": dump_filename,
                "NUITKA_PROJECT_DIR": project_dir,
                "PYTHONPATH": os.pathsep.join(
                    [mock_site_packages_path]
                    + [
                        candidate
                        for candidate in sys.path
                        if os.path.dirname(candidate) in getSitePackageCandidateNames()
                    ]
                ),
            }
        ):
            with withDirectoryChange(temp_dir):
                command = (
                    sys.executable,
                    "-W",
                    "ignore",
                    "-m",
                    "build",
                    "--no-isolation",
                    "--skip-dependency-check",
                    project_dir,
                )

                stdout, stderr, exit_code = executeProcess(
                    command,
                    stdin=False,
                )

        # The build is expected to fail with exit code 1 because we exit(0)
        # without producing a wheel, causing build frontend to complain. But if
        # we have the dump file, we are good.
        if exit_code == 0 or not os.path.exists(dump_filename):
            reportBuildError(logger, "python -m build", command, stdout, stderr)

        config = loadJsonFromFilename(dump_filename)

        package_dir = config.get("package_dir")
        if package_dir and "" in package_dir:
            addMainScriptDirectory(os.path.join(os.getcwd(), package_dir.get("")))
        else:
            addMainScriptDirectory(os.getcwd())

    setProjectName(config.get("project_name"))

    return config.get("arguments", [])


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
dblite.py module contributed by Ralf W. Grosse-Kunstleve.
Extended for Unicode by Steven Knight.

This is a very simple-minded "database" used for saved signature
information, with an interface modeled on the Python dbm database
interface module.
"""

import io
import os
import pickle
import shutil
import time

from SCons.compat import PICKLE_PROTOCOL

KEEP_ALL_FILES = False
IGNORE_CORRUPT_DBFILES = False


def corruption_warning(filename):
    """Local warning for corrupt db.

    Used for self-tests. SCons overwrites this with a
    different warning function in SConsign.py.
    """
    print("Warning: Discarding corrupt database:", filename)


DBLITE_SUFFIX = ".dblite"
TMP_SUFFIX = ".tmp"


class _Dblite:
    """Lightweight signature database class.

    Behaves like a dict when in memory, loads from a pickled disk
    file on open and writes back out to it on close.

    Open the database file using a path derived from *file_base_name*.
    The optional *flag* argument can be:

    +---------+---------------------------------------------------+
    | Value   | Meaning                                           |
    +=========+===================================================+
    | ``'r'`` | Open existing database for reading only (default) |
    +---------+---------------------------------------------------+
    | ``'w'`` | Open existing database for reading and  writing   |
    +---------+---------------------------------------------------+
    | ``'c'`` | Open database for reading and writing, creating   |
    |         | it if it doesn't exist                            |
    +---------+---------------------------------------------------+
    | ``'n'`` | Always create a new, empty database, open for     |
    |         | reading and writing                               |
    +---------+---------------------------------------------------+

    The optional *mode* argument is the POSIX mode of the file, used only
    when the database has to be created.  It defaults to octal ``0o666``.
    """

    # Because open() is defined at module level, overwriting builtin open
    # in the scope of this module, we use io.open to avoid ambiguity.
    _open = staticmethod(io.open)

    # we need to squirrel away references to functions from various modules
    # that we'll use when sync() is called: this may happen at Python
    # teardown time (we call it from our __del__), and the global module
    # references themselves may already have been rebound to None.
    _pickle_dump = staticmethod(pickle.dump)
    _pickle_protocol = PICKLE_PROTOCOL
    try:
        _os_chown = staticmethod(os.chown)
    except AttributeError:
        _os_chown = None
    _os_replace = staticmethod(os.replace)
    _os_chmod = staticmethod(os.chmod)
    _shutil_copyfile = staticmethod(shutil.copyfile)
    _time_time = staticmethod(time.time)

    def __init__(self, file_base_name, flag='r', mode=0o666):
        assert flag in ("r", "w", "c", "n")

        base, ext = os.path.splitext(file_base_name)
        if ext == DBLITE_SUFFIX:
            # There's already a suffix on the file name, don't add one.
            self._file_name = file_base_name
            self._tmp_name = base + TMP_SUFFIX
        else:
            self._file_name = file_base_name + DBLITE_SUFFIX
            self._tmp_name = file_base_name + TMP_SUFFIX

        self._flag = flag
        self._mode = mode
        self._dict = {}
        self._needs_sync = False

        if self._os_chown is not None and 0 in (os.geteuid(), os.getegid()):
            # running as root; chown back to current owner/group when done
            try:
                statinfo = os.stat(self._file_name)
                self._chown_to = statinfo.st_uid
                self._chgrp_to = statinfo.st_gid
            except OSError:
                # db file doesn't exist yet.
                # Check os.environ for SUDO_UID, use if set
                self._chown_to = int(os.environ.get('SUDO_UID', -1))
                self._chgrp_to = int(os.environ.get('SUDO_GID', -1))
        else:
            self._chown_to = -1  # don't chown
            self._chgrp_to = -1  # don't chgrp

        if self._flag == "n":
            with io.open(self._file_name, "wb", opener=self.opener):
                return  # just make sure it exists
        else:
            # We only need the disk file to slurp in the data.  Updates are
            # handled on close, db is mainained only in memory until then.
            try:
                with io.open(self._file_name, "rb") as f:
                    p = f.read()
            except OSError as e:
                # an error for file not to exist, unless flag is create
                if self._flag != "c":
                    raise e
                with io.open(self._file_name, "wb", opener=self.opener):
                    return  # just make sure it exists
            if len(p) > 0:
                try:
                    self._dict = pickle.loads(p, encoding='bytes')
                except (
                    pickle.UnpicklingError,
                    # Python3 docs:
                    # Note that other exceptions may also be raised during
                    # unpickling, including (but not necessarily limited to)
                    # AttributeError, EOFError, ImportError, and IndexError.
                    AttributeError,
                    EOFError,
                    ImportError,
                    IndexError,
                ):
                    if IGNORE_CORRUPT_DBFILES:
                        corruption_warning(self._file_name)
                    else:
                        raise

    def opener(self, path, flags):
        """Database open helper when creation may be needed.

        The high-level Python open() function cannot specify a file mode
        for creation. Using this as the opener with the saved mode lets
        us do that.
        """
        return os.open(path, flags, mode=self._mode)

    def close(self):
        if self._needs_sync:
            self.sync()

    def __del__(self):
        self.close()

    def sync(self):
        """Flush the database to disk.

        This routine *must* succeed, since the in-memory and on-disk
        copies are out of sync as soon as we do anything that changes
        the in-memory version. Thus, to be cautious, flush to a
        temporary file and then move it over with some error handling.
        """
        self._check_writable()
        with self._open(self._tmp_name, "wb", opener=self.opener) as f:
            self._pickle_dump(self._dict, f, self._pickle_protocol)

        try:
            self._os_replace(self._tmp_name, self._file_name)
        except PermissionError:
            # If we couldn't replace due to perms, try to change and retry.
            # This is mainly for Windows - on POSIX the file permissions
            # don't matter, the os.replace would have worked anyway.
            # We're giving up if the retry fails, just let the Python
            # exception abort us.
            try:
                self._os_chmod(self._file_name, 0o777)
            except PermissionError:
                pass
            self._os_replace(self._tmp_name, self._file_name)

        if (
            self._os_chown is not None and self._chown_to > 0
        ):  # don't chown to root or -1
            try:
                self._os_chown(self._file_name, self._chown_to, self._chgrp_to)
            except OSError:
                pass

        self._needs_sync = False
        if KEEP_ALL_FILES:
            self._shutil_copyfile(
                self._file_name, f"{self._file_name}_{int(self._time_time())}"
            )

    def _check_writable(self):
        if self._flag == "r":
            raise OSError(f"Read-only database: {self._file_name}")

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        self._check_writable()

        if not isinstance(key, str):
            raise TypeError(f"key `{key}' must be a string but is {type(key)}")

        if not isinstance(value, bytes):
            raise TypeError(f"value `{value}' must be bytes but is {type(value)}")

        self._dict[key] = value
        self._needs_sync = True

    def __delitem__(self, key):
        del self._dict[key]

    def keys(self):
        return self._dict.keys()

    def items(self):
        return self._dict.items()

    def values(self):
        return self._dict.values()

    __iter__ = keys

    def __contains__(self, key):
        return key in self._dict

    def __len__(self):
        return len(self._dict)


def open(file, flag="r", mode = 0o666):  # pylint: disable=redefined-builtin
    return _Dblite(file, flag, mode)


def _exercise():
    db = open("tmp", "n")
    assert len(db) == 0
    db["foo"] = b"bar"
    assert db["foo"] == b"bar"
    db.sync()

    db = open("tmp", "c")
    assert len(db) == 1, len(db)
    assert db["foo"] == b"bar"
    db["bar"] = b"foo"
    assert db["bar"] == b"foo"
    db.sync()

    db = open("tmp")
    assert len(db) == 2, len(db)
    assert db["foo"] == b"bar"
    assert db["bar"] == b"foo"
    try:
        db.sync()
    except OSError as e:
        assert str(e) == "Read-only database: tmp.dblite"
    else:
        raise RuntimeError("IOError expected.")
    db = open("tmp", "w")
    assert len(db) == 2, len(db)
    db["ping"] = b"pong"
    db.sync()

    try:
        db[(1, 2)] = "tuple"
    except TypeError as e:
        assert str(e) == "key `(1, 2)' must be a string but is <class 'tuple'>", str(e)
    else:
        raise RuntimeError("TypeError exception expected")

    try:
        db["list"] = [1, 2]
    except TypeError as e:
        assert str(e) == "value `[1, 2]' must be bytes but is <class 'list'>", str(e)
    else:
        raise RuntimeError("TypeError exception expected")

    db = open("tmp")
    assert len(db) == 3, len(db)

    db = open("tmp", "n")
    assert len(db) == 0, len(db)
    _Dblite._open("tmp.dblite", "w")

    db = open("tmp")
    _Dblite._open("tmp.dblite", "w").write("x")
    try:
        db = open("tmp")
    except pickle.UnpicklingError:
        pass
    else:
        raise RuntimeError("pickle exception expected.")

    global IGNORE_CORRUPT_DBFILES
    IGNORE_CORRUPT_DBFILES = True
    db = open("tmp")
    assert len(db) == 0, len(db)
    os.unlink("tmp.dblite")
    try:
        db = open("tmp", "w")
    except OSError as e:
        assert str(e) == "[Errno 2] No such file or directory: 'tmp.dblite'", str(e)
    else:
        raise RuntimeError("IOError expected.")

    print("Completed _exercise()")


if __name__ == "__main__":
    _exercise()

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from SCons.Errors import UserError
from SCons.Tool import createSharedLibBuilder
from SCons.Util import CLVar
from . import lib_emitter, EmitLibSymlinks, StringizeLibSymlinks


def shlib_symlink_emitter(target, source, env, **kw):
    verbose = False

    if "variable_prefix" in kw:
        var_prefix = kw["variable_prefix"]
    else:
        var_prefix = "SHLIB"

    do_symlinks = env.subst("$%sNOVERSIONSYMLINKS" % var_prefix)
    if do_symlinks in ["1", "True", "true", True]:
        return target, source

    shlibversion = env.subst("$%sVERSION" % var_prefix)
    if shlibversion:
        if verbose:
            print("shlib_symlink_emitter: %sVERSION=%s" % (var_prefix, shlibversion))

        libnode = target[0]
        shlib_soname_symlink = env.subst(
            "$%s_SONAME_SYMLINK" % var_prefix, target=target, source=source
        )
        shlib_noversion_symlink = env.subst(
            "$%s_NOVERSION_SYMLINK" % var_prefix, target=target, source=source
        )

        if verbose:
            print("shlib_soname_symlink    :%s" % shlib_soname_symlink)
            print("shlib_noversion_symlink :%s" % shlib_noversion_symlink)
            print("libnode                 :%s" % libnode)

        shlib_soname_symlink = env.File(shlib_soname_symlink)
        shlib_noversion_symlink = env.File(shlib_noversion_symlink)

        symlinks = []
        if shlib_soname_symlink != libnode:
            # If soname and library name machine, don't symlink them together
            symlinks.append((env.File(shlib_soname_symlink), libnode))

        symlinks.append((env.File(shlib_noversion_symlink), libnode))

        if verbose:
            print(
                "_lib_emitter: symlinks={!r}".format(
                    ", ".join(
                        ["%r->%r" % (k, v) for k, v in StringizeLibSymlinks(symlinks)]
                    )
                )
            )

        if symlinks:
            # This does the actual symlinking
            EmitLibSymlinks(env, symlinks, target[0])

            # This saves the information so if the versioned shared library is installed
            # it can faithfully reproduce the correct symlinks
            target[0].attributes.shliblinks = symlinks

    return target, source


def _soversion(target, source, env, for_signature):
    """Function to determine what to use for SOVERSION"""

    if "SOVERSION" in env:
        return ".$SOVERSION"
    elif "SHLIBVERSION" in env:
        shlibversion = env.subst("$SHLIBVERSION")
        # We use only the most significant digit of SHLIBVERSION
        return "." + shlibversion.split(".")[0]
    else:
        return ""


def _soname(target, source, env, for_signature):
    if "SONAME" in env:
        # Now verify that SOVERSION is not also set as that is not allowed
        if "SOVERSION" in env:
            raise UserError(
                "Ambiguous library .so naming, both SONAME: %s and SOVERSION: %s are defined. "
                "Only one can be defined for a target library."
                % (env["SONAME"], env["SOVERSION"])
            )
        return "$SONAME"
    else:
        return "$SHLIBPREFIX$_get_shlib_stem${SHLIBSUFFIX}$_SHLIBSOVERSION"


def _get_shlib_stem(target, source, env, for_signature):
    """Get the base name of a shared library.

    Args:
        target: target node containing the lib name
        source: source node, not used
        env: environment context for running subst
        for_signature: whether this is being done for signature generation

    Returns:
        the library name without prefix/suffix
    """
    verbose = False

    target_name = str(target.name)
    shlibprefix = env.subst("$SHLIBPREFIX")
    shlibsuffix = env.subst("$_SHLIBSUFFIX")

    if verbose and not for_signature:
        print(
            "_get_shlib_stem: target_name:%s shlibprefix:%s shlibsuffix:%s"
            % (target_name, shlibprefix, shlibsuffix)
        )


    if shlibsuffix and target_name.endswith(shlibsuffix):
        target_name = target_name[: -len(shlibsuffix)]

    if shlibprefix and target_name.startswith(shlibprefix):
        # skip pathological case were target _is_ the prefix
        if target_name != shlibprefix:
            target_name = target_name[len(shlibprefix) :]


    if verbose and not for_signature:
        print("_get_shlib_stem: target_name:%s AFTER" % (target_name,))

    return target_name


def _get_shlib_dir(target, source, env, for_signature):
    """Get the directory the shared library is in.

    Args:
        target: target node
        source: source node, not used
        env: environment context, not used
        for_signature: whether this is being done for signature generation

    Returns:
        the directory the library will be in (empty string if '.')
    """
    verbose = False

    if target.dir and str(target.dir) != ".":
        if verbose:
            print("_get_shlib_dir: target.dir:%s" % target.dir)

        return "%s/" % str(target.dir)
    else:
        return ""


def setup_shared_lib_logic(env):
    """Initialize an environment for shared library building.

    Args:
        env: environment to set up
    """
    createSharedLibBuilder(env)

    env["_get_shlib_stem"] = _get_shlib_stem
    env["_get_shlib_dir"] = _get_shlib_dir
    env["_SHLIBSOVERSION"] = _soversion
    env["_SHLIBSONAME"] = _soname

    env["SHLIBNAME"] = "${_get_shlib_dir}${SHLIBPREFIX}$_get_shlib_stem${_SHLIBSUFFIX}"

    # This is the non versioned shlib filename
    # If SHLIBVERSION is defined then this will symlink to $SHLIBNAME
    env["SHLIB_NOVERSION_SYMLINK"] = "${_get_shlib_dir}${SHLIBPREFIX}$_get_shlib_stem${SHLIBSUFFIX}"

    # This is the sonamed file name
    # If SHLIBVERSION is defined then this will symlink to $SHLIBNAME
    env["SHLIB_SONAME_SYMLINK"] = "${_get_shlib_dir}$_SHLIBSONAME"

    # Note this is gnu style
    env["SHLIBSONAMEFLAGS"] = "-Wl,-soname=$_SHLIBSONAME"
    env["_SHLIBVERSION"] = "${SHLIBVERSION and '.'+SHLIBVERSION or ''}"
    env["_SHLIBVERSIONFLAGS"] = "$SHLIBVERSIONFLAGS -Wl,-soname=$_SHLIBSONAME"

    env["SHLIBEMITTER"] = [lib_emitter, shlib_symlink_emitter]

    # If it's already set, then don't overwrite.
    env["SHLIBPREFIX"] = env.get('SHLIBPREFIX', "lib")
    env["_SHLIBSUFFIX"] = "${SHLIBSUFFIX}${_SHLIBVERSION}"

    env["SHLINKFLAGS"] = CLVar("$LINKFLAGS -shared")

    env["SHLINKCOM"] = "$SHLINK -o $TARGET $SHLINKFLAGS $__SHLIBVERSIONFLAGS $__RPATH $SOURCES $_LIBDIRFLAGS $_LIBFLAGS"

    env["SHLINK"] = "$LINK"



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Code generation for networkx module specific stuff."""

from .BuiltinCodes import getBuiltinCallViaSpecCode
from .ImportCodes import getImportModuleNameHardCode
from .JitCodes import addUncompiledFunctionSourceDict


def generateNetworkxUtilsDecoratorsArgmapCallCode(to_name, expression, emit, context):
    """This is for networkx.utils.decorators.argmap calls."""

    # TODO: Have global cached forms of hard attribute lookup results too.
    argmap_class_name = context.allocateTempName("argmap_class", unique=True)

    getImportModuleNameHardCode(
        to_name=argmap_class_name,
        module_name="networkx.utils.decorators",
        import_name="argmap",
        needs_check=False,
        emit=emit,
        context=context,
    )

    addUncompiledFunctionSourceDict(func_value=expression.subnode_func, context=context)

    getBuiltinCallViaSpecCode(
        spec=expression.spec,
        called_name=argmap_class_name,
        to_name=to_name,
        expression=expression,
        emit=emit,
        context=context,
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Access package data."""

import os
import pkgutil

from nuitka.importing.Importing import locateModule
from nuitka.PythonVersions import python_version

from .FileOperations import getFileContents


def getPackageData(package_name, resource):
    """Get the package data, but without loading the code, i.e. we try and avoids its loader.

    If it's absolutely necessary, we fallback to "pkgutil.get_data" but
    only if we absolutely cannot find it easily.
    """
    package_directory = locateModule(package_name, None, 0)[1]

    if package_directory is not None:
        resource_filename = os.path.join(package_directory, resource)

        if os.path.exists(resource_filename):
            return getFileContents(resource_filename, mode="rb")

    # Try importlib.resources next for a more standard approach.
    try:
        if python_version >= 0x370:
            import importlib.resources as importlib_resources

            return importlib_resources.read_binary(package_name.asString(), resource)
        elif python_version >= 0x350:
            # Use backport or older importlib.resources
            import importlib.resources as importlib_resources

            with importlib_resources.open_binary(
                package_name.asString(), resource
            ) as fp:
                return fp.read()
    except (ImportError, FileNotFoundError, OSError, TypeError):
        pass

    # Fallback to pkgutil if all else fails, this however may import
    # the package, which is not a good idea at all.
    return pkgutil.get_data(package_name.asString(), resource)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Time taking.

Mostly for measurements of Nuitka of itself, e.g. how long did it take to
call an external tool.
"""

from contextlib import contextmanager

from nuitka.__past__ import StringIO, perf_counter, process_time
from nuitka.Tracing import general

from .Profiling import PerfCounters, hasPerfProfilingSupport

has_perf_counters = hasPerfProfilingSupport()


class StopWatchWallClockBase(object):
    __slots__ = ("start_time", "end_time", "perf_counters")

    # For overload, pylint: disable=not-callable
    timer = None

    def __init__(self, use_perf_counters=False):
        self.start_time = None
        self.end_time = None

        self.perf_counters = PerfCounters() if use_perf_counters else None

    def start(self):
        if self.perf_counters is not None:
            self.perf_counters.start()
        self.start_time = self.timer()

    def restart(self):
        self.start()

    def end(self):
        self.end_time = self.timer()

        if self.perf_counters is not None:
            self.perf_counters.stop()

    stop = end

    def getDelta(self):
        if self.end_time is not None:
            return self.end_time - self.start_time
        else:
            return self.timer() - self.start_time

    def getPerfCounters(self):
        if self.perf_counters is not None:
            return self.perf_counters.getValues()
        else:
            return None, None


class StopWatchWallClock(StopWatchWallClockBase):
    timer = perf_counter


class StopWatchProcessClock(StopWatchWallClockBase):
    timer = process_time


class TimerReport(object):
    """Timer that reports how long things took.

    Mostly intended as a wrapper for external process calls.
    """

    __slots__ = (
        "message",
        "decider",
        "logger",
        "timer",
        "min_report_time",
        "include_sleep_time",
        "use_perf_counters",
    )

    def __init__(
        self,
        message,
        logger=None,
        decider=True,
        min_report_time=None,
        include_sleep_time=True,
        use_perf_counters=None,
    ):
        self.message = message

        # Shortcuts.
        if decider is True:
            decider = lambda: 1
        elif decider is False:
            decider = lambda: 0

        if logger is None:
            logger = general

        self.logger = logger
        self.decider = decider
        self.min_report_time = min_report_time

        self.timer = None
        self.include_sleep_time = include_sleep_time

        if use_perf_counters is None:
            use_perf_counters = not self.include_sleep_time

        # They might not be allowed.
        self.use_perf_counters = use_perf_counters and has_perf_counters

    def getTimer(self):
        return self.timer

    def __enter__(self):
        stop_stop_class = (
            StopWatchWallClock if self.include_sleep_time else StopWatchProcessClock
        )
        self.timer = stop_stop_class(self.use_perf_counters)

        self.timer.start()

        return self.timer

    def __exit__(self, exception_type, exception_value, exception_tb):
        self.timer.end()

        delta_time = self.timer.getDelta()

        # Check if its above the provided limit.
        above_threshold = (
            self.min_report_time is None or delta_time >= self.min_report_time
        )

        if exception_type is None and above_threshold and self.decider():
            self.logger.info(self.message % self.timer.getDelta(), keep_format=True)


@contextmanager
def withProfiling(name, logger, enabled):
    if enabled:
        import cProfile
        import pstats

        from nuitka.options.Options import getOutputPath

        pr = cProfile.Profile(timer=process_time)
        pr.enable()

        yield

        pr.disable()

        s = StringIO()
        ps = pstats.Stats(pr, stream=s).sort_stats(pstats.SortKey.CUMULATIVE)

        ps.print_stats()
        for line in s.getvalue().splitlines():
            logger.info(line)

        profile_filename = getOutputPath(name + ".prof")

        pr.dump_stats(profile_filename)
        logger.info("Profiling data for '%s' saved to '%s'." % (name, profile_filename))
    else:
        yield


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Jorj McKie, mailto:<jorj.x.mckie@outlook.de> find license text at end of file


"""Details see below in class definition."""

import os
import sys

from nuitka.options.Options import isStandaloneMode, shallCreateAppBundle
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.PythonFlavors import isHomebrewPython
from nuitka.PythonVersions import getSystemPrefixPath, getTkInterVersion
from nuitka.utils.Utils import isMacOS, isWin32Windows

# spell-checker: ignore tcltk,tcltest


def _isTkInterModule(module):
    full_name = module.getFullName()
    return full_name in ("Tkinter", "tkinter", "PySimpleGUI", "PySimpleGUI27")


def _getHomebrewPrefix(logger):
    result = os.path.normpath(
        os.path.join(getSystemPrefixPath(), "..", "..", "..", "..", "..", "..", "..")
    )

    if not os.path.isdir(result):
        logger.sysexit("Error, failed to determine Homebrew prefix, report this bug.")

    return result


class NuitkaPluginTkinter(NuitkaPluginBase):
    """This class represents the main logic of the TkInter plugin.

     This is a plug-in to make programs work well in standalone mode which are using tkinter.
     These programs require the presence of certain libraries written in the TCL language.
     On Windows platforms, and even on Linux, the existence of these libraries cannot be
     assumed. We therefore

     1. Copy the TCL libraries as sub-folders to the program's dist folder
     2. Redirect the program's tkinter requests to these library copies. This is
        done by setting appropriate variables in the os.environ dictionary.
        Tkinter will use these variable value to locate the library locations.

     Each time before the program issues an import to a tkinter module, we make
     sure, that the TCL environment variables are correctly set.

    Notes:
         You can enforce using a specific TCL folder by using TCL_LIBRARY
         and a Tk folder by using TK_LIBRARY, but that ought to normally
         not be necessary.
    """

    plugin_name = "tk-inter"  # Nuitka knows us by this name
    plugin_desc = "Required by 'tkinter' package."
    plugin_category = "package-support"

    # Automatically suppress detectors for any other toolkit
    plugin_gui_toolkit = True

    # Only used in control tags
    binding_name = "tkinter"

    def __init__(self, tcl_library_dir, tk_library_dir):
        self.tcl_library_dir = tcl_library_dir
        self.tk_library_dir = tk_library_dir

        # ensure one-time action, we deal with several names for the execution,
        # yet we only want to do it once.
        self.files_copied = False

        self.tk_inter_version = getTkInterVersion()

        if self.tk_inter_version is None:
            self.sysexit("Error, it seems 'tk-inter' is not installed.")

        # Only ever saw these in use, report if there are more.
        if self.tk_inter_version not in ("8.5", "8.6", "9.0"):
            self.sysexit(
                """\
Error, it seems 'tk-inter' has an unsupported version '%s'. \
Please report as a issue."""
                % self.tk_inter_version
            )

        return None

    @classmethod
    def isRelevant(cls):
        """This method is called one time only to check, whether the plugin might make sense at all.

        Returns:
            True if this is a standalone, else False.
        """
        return isStandaloneMode()

    def createPreModuleLoadCode(self, module):
        """This method is called with a module that will be imported.

        Notes:
            If the word "tkinter" occurs in its full name, we know that the correct
            setting of the TCL environment must be ensured before this happens.

        Args:
            module: the module object
        Returns:
            Code to insert and None (tuple)
        """
        # only insert code for tkinter related modules
        if _isTkInterModule(module):
            # The following code will be executed before importing the module.
            # If required we set the respective environment values.
            code = r"""
import os
os.environ["TCL_LIBRARY"] = os.path.join(__nuitka_binary_dir, "%(tcl_target_path)s")
os.environ["TK_LIBRARY"] = os.path.join(__nuitka_binary_dir, "%(tk_target_path)s")""" % {
                "tcl_target_path": self.getTclTargetPath(),
                "tk_target_path": self.getTkTargetPath(),
            }

            return code, "Need to make sure we set environment variables for TCL."

    @classmethod
    def addPluginCommandLineOptions(cls, group):
        group.add_option(
            "--tk-library-dir",
            action="store",
            dest="tk_library_dir",
            default=None,
            help="""\
The Tk library dir. Nuitka is supposed to automatically detect it, but you can
override it here. Default is automatic detection.""",
        )

        group.add_option(
            "--tcl-library-dir",
            action="store",
            dest="tcl_library_dir",
            default=None,
            help="""\
The Tcl library dir. See comments for Tk library dir.""",
        )

    def _getTclCandidatePaths(self):
        # Check typical locations of the dirs
        yield os.getenv("TCL_LIBRARY")

        # Inside the Python install, esp. on Windows.
        for sys_prefix_path in (sys.prefix, getSystemPrefixPath()):
            yield os.path.join(sys_prefix_path, "tcl", "tcl%s" % self.tk_inter_version)
            yield os.path.join(sys_prefix_path, "lib", "tcl%s" % self.tk_inter_version)

            # Newer Anaconda.
            yield os.path.join(
                sys_prefix_path, "Library", "lib", "tcl%s" % self.tk_inter_version
            )

        # System installs on non-Windows
        if not isWin32Windows():
            yield "/usr/share/tcltk/tcl%s" % self.tk_inter_version
            yield "/usr/share/tcl%s" % self.tk_inter_version
            yield "/usr/lib64/tcl/tcl%s" % self.tk_inter_version
            yield "/usr/lib/tcl%s" % self.tk_inter_version

        if isHomebrewPython():
            yield os.path.normpath(
                os.path.join(
                    _getHomebrewPrefix(self),
                    "lib",
                    "tcl%s" % self.tk_inter_version,
                )
            )

            # Homebrew is compiled to think it's 8.6, but it might actually
            # be the version 9.
            yield os.path.normpath(
                os.path.join(
                    _getHomebrewPrefix(self),
                    "lib",
                    "tcl9",
                )
            )

        if isMacOS():
            yield os.path.normpath(
                os.path.join(
                    getSystemPrefixPath(),
                    "Frameworks",
                    "Tcl.framework",
                    "Resources",
                    "Scripts",
                )
            )

    def _getTkCandidatePaths(self):
        yield os.getenv("TK_LIBRARY")

        for sys_prefix_path in (sys.prefix, getSystemPrefixPath()):
            yield os.path.join(sys_prefix_path, "tcl", "tk%s" % self.tk_inter_version)
            yield os.path.join(sys_prefix_path, "lib", "tk%s" % self.tk_inter_version)

            # Newer Anaconda.
            yield os.path.join(
                sys_prefix_path, "Library", "lib", "tk%s" % self.tk_inter_version
            )

        if not isWin32Windows():
            yield "/usr/share/tcltk/tk%s" % self.tk_inter_version
            yield "/usr/share/tk%s" % self.tk_inter_version
            yield "/usr/lib64/tcl/tk%s" % self.tk_inter_version
            yield "/usr/lib/tk%s" % self.tk_inter_version

        if isHomebrewPython():
            yield os.path.normpath(
                os.path.join(
                    _getHomebrewPrefix(self),
                    "lib",
                    "tk%s" % self.tk_inter_version,
                )
            )

            if self.tk_inter_version == "8.6":
                # Homebrew is compiled to think it's 8.6, but it might actually
                # be the version 9.
                yield os.path.normpath(
                    os.path.join(
                        _getHomebrewPrefix(self),
                        "lib",
                        "tk9.0",
                    )
                )

        if isMacOS():
            yield os.path.normpath(
                os.path.join(
                    getSystemPrefixPath(),
                    "Frameworks",
                    "Tk.framework",
                    "Resources",
                    "Scripts",
                )
            )

    @staticmethod
    def getTclTargetPath():
        if isMacOS():
            return "tcl-files"
        else:
            return "tcl"

    @staticmethod
    def getTkTargetPath():
        if isMacOS():
            return "tk-files"
        else:
            return "tk"

    def considerDataFiles(self, module):
        """Provide TCL libraries to the dist folder.

        Notes:
            We will provide the copy the TCL/TK directories to the program's root directory,
            that might be shiftable with some work.

        Args:
            module: the module in question, maybe ours

        Yields:
            IncludedDataFile objects.
        """

        if not _isTkInterModule(module) or self.files_copied:
            return

        tcl_library_dir = self.tcl_library_dir
        if tcl_library_dir is None:
            for tcl_library_dir in self._getTclCandidatePaths():
                if tcl_library_dir is not None and os.path.exists(
                    os.path.join(tcl_library_dir, "init.tcl")
                ):
                    break

        if tcl_library_dir is None or not os.path.exists(tcl_library_dir):
            self.sysexit(
                """\
Could not find Tcl, you might need to use '--tcl-library-dir' and if \
that works, report a bug so it can be added to Nuitka."""
            )

        tk_library_dir = self.tk_library_dir
        if tk_library_dir is None:
            for tk_library_dir in self._getTkCandidatePaths():
                if tk_library_dir is not None and os.path.exists(
                    os.path.join(tk_library_dir, "dialog.tcl")
                ):
                    break

        if tk_library_dir is None or not os.path.exists(tk_library_dir):
            self.sysexit(
                """\
Could not find Tk, you might need to use '--tk-library-dir' and if \
that works, report a bug."""
            )

        # survived the above, now do provide the locations
        yield self.makeIncludedDataDirectory(
            source_path=tk_library_dir,
            dest_path=self.getTkTargetPath(),
            reason="Tk needed for tkinter usage",
            ignore_dirs=("demos",),
            tags="tk",
        )
        yield self.makeIncludedDataDirectory(
            source_path=tcl_library_dir,
            ignore_dirs=(("opt0.4", "http1.0") if shallCreateAppBundle() else ()),
            # TODO: Not very version robust, may we ought to
            # become able to ignore files by pattern.
            ignore_filenames=(
                "tcltest-2.3.6.tm",
                "tcltest-2.3.8.tm",
                "tcltest-2.4.0.tm",
                "tcltest-2.5.0.tm",
                "tcltest-2.5.3.tm",
                "tcltest-2.5.5.tm",
                "tcltest-2.5.8.tm",
            ),
            dest_path=self.getTclTargetPath(),
            reason="Tcl needed for tkinter usage",
            tags="tcl",
        )

        if isWin32Windows():
            yield self.makeIncludedDataDirectory(
                source_path=os.path.join(tcl_library_dir, "..", "tcl8"),
                dest_path="tcl8",
                reason="Tcl needed for tkinter usage",
                tags="tcl",
            )

        self.files_copied = True

    def onModuleCompleteSet(self, module_set):
        if str is bytes:
            plugin_binding_name = "Tkinter"
        else:
            plugin_binding_name = "tkinter"

        self.onModuleCompleteSetGUI(
            module_set=module_set, plugin_binding_name=plugin_binding_name
        )


class NuitkaPluginDetectorTkinter(NuitkaPluginBase):
    """Used only if plugin is not activated.

    Notes:
        We are given the chance to issue a warning if we think we may be required.
    """

    detector_for = NuitkaPluginTkinter

    @classmethod
    def isRelevant(cls):
        """This method is called one time only to check, whether the plugin might make sense at all.

        Returns:
            True if this is a standalone compilation on Windows, else False.
        """
        return isStandaloneMode()

    def checkModuleSourceCode(self, module_name, source_code):
        """This method checks the source code

        Notes:
            We only use it to check whether this is the main module, and whether
            it contains the keyword "tkinter".
            We assume that the main program determines whether tkinter is used.
            References by dependent or imported modules are assumed irrelevant.

        Args:
            module_name: the name of the module
            source_code: the module's source code

        Returns:
            None
        """
        if module_name == "__main__":
            for line in source_code.splitlines():
                # Ignore comments.
                if "#" in line:
                    line = line[: line.find("#")]

                if "tkinter" in line or "Tkinter" in line:
                    self.warnUnusedPlugin("Tkinter needs TCL included.")
                    break

    def getReportData(self):
        return {
            "tk_inter_version": self.tk_inter_version,
            "tcl_library_dir": self.tcl_library_dir,
            "tk_library_dir": self.tk_library_dir,
        }


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Code generation for standard CPython/API calls.

This is generic stuff, geared at calling functions that accept Python objects
and return Python objects. As these all work in a similar way, it makes sense
to concentrate the best way to do to make those calls here.

Also, many Nuitka helper codes turn out to be very similar to Python C/API
and then can use the same code.
"""

from .CodeHelpers import generateExpressionCode
from .ErrorCodes import (
    getErrorExitBoolCode,
    getErrorExitCode,
    getReleaseCode,
    getReleaseCodes,
)


def makeArgDescFromExpression(expression):
    """Helper for providing arg_desc consistently for generateCAPIObject methods."""

    return tuple(
        (child_name + "_value", child_value)
        for child_name, child_value in expression.getVisitableNodesNamed()
    )


def generateCAPIObjectCodeCommon(
    to_name,
    capi,
    tstate,
    arg_desc,
    may_raise,
    conversion_check,
    ref_count,
    source_ref,
    emit,
    context,
    none_null=False,
):
    arg_names = []

    if tstate:
        arg_names.append("tstate")

    for arg_name, arg_expression in arg_desc:
        if arg_expression is None and none_null:
            arg_names.append("NULL")
        else:
            arg_name = context.allocateTempName(arg_name)

            generateExpressionCode(
                to_name=arg_name,
                expression=arg_expression,
                emit=emit,
                context=context,
            )

            arg_names.append(arg_name)

    context.setCurrentSourceCodeReference(source_ref)

    getCAPIObjectCode(
        to_name=to_name,
        capi=capi,
        arg_names=arg_names,
        may_raise=may_raise,
        conversion_check=conversion_check,
        ref_count=ref_count,
        emit=emit,
        context=context,
    )


def generateCAPIObjectCode(
    to_name,
    capi,
    tstate,
    arg_desc,
    may_raise,
    conversion_check,
    source_ref,
    emit,
    context,
    none_null=False,
):
    generateCAPIObjectCodeCommon(
        to_name=to_name,
        capi=capi,
        tstate=tstate,
        arg_desc=arg_desc,
        may_raise=may_raise,
        conversion_check=conversion_check,
        ref_count=1,
        source_ref=source_ref,
        emit=emit,
        context=context,
        none_null=none_null,
    )


def generateCAPIObjectCode0(
    to_name,
    capi,
    tstate,
    arg_desc,
    may_raise,
    conversion_check,
    source_ref,
    emit,
    context,
    none_null=False,
):
    generateCAPIObjectCodeCommon(
        to_name=to_name,
        capi=capi,
        tstate=tstate,
        arg_desc=arg_desc,
        may_raise=may_raise,
        conversion_check=conversion_check,
        ref_count=0,
        source_ref=source_ref,
        emit=emit,
        context=context,
        none_null=none_null,
    )


def getCAPIObjectCode(
    to_name, capi, arg_names, may_raise, conversion_check, ref_count, emit, context
):
    release_names = tuple(
        arg_name for arg_name in arg_names if arg_name not in ("NULL", "tstate")
    )

    if to_name is not None:
        # TODO: Use context manager here too.
        if to_name.c_type == "PyObject *":
            value_name = to_name
        else:
            value_name = context.allocateTempName("capi_result")

        emit(
            "%s = %s(%s);"
            % (value_name, capi, ", ".join(str(arg_name) for arg_name in arg_names))
        )

        getErrorExitCode(
            check_name=value_name,
            release_names=release_names,
            needs_check=may_raise,
            emit=emit,
            context=context,
        )

        if ref_count:
            context.addCleanupTempName(value_name)

        if to_name is not value_name:
            to_name.getCType().emitAssignConversionCode(
                to_name=to_name,
                value_name=value_name,
                needs_check=conversion_check,
                emit=emit,
                context=context,
            )

            if ref_count:
                getReleaseCode(value_name, emit, context)
    else:
        if may_raise:
            res_name = context.getIntResName()

            emit(
                "%s = %s(%s);"
                % (res_name, capi, ", ".join(str(arg_name) for arg_name in arg_names))
            )

            getErrorExitBoolCode(
                condition="%s == -1" % res_name,
                release_names=release_names,
                emit=emit,
                context=context,
            )
        else:
            emit("%s(%s);" % (capi, ", ".join(str(arg_name) for arg_name in arg_names)))

            getReleaseCodes(release_names, emit, context)

        assert not ref_count


def getReferenceExportCode(base_name, emit, context):
    if not context.needsCleanup(base_name):
        emit("Py_INCREF(%s);" % base_name)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""The SCons Warnings framework.

Enables issuing warnings in situations where it is useful to alert
the user of a condition that does not warrant raising an exception
that could terminate the program.

A new warning class should inherit (perhaps indirectly) from one of
two base classes: :exc:`SConsWarning` or :exc:`WarningOnByDefault`,
which are the same except warnings derived from the latter will start
out in an enabled state. Enabled warnings cause a message to be
printed when called, disabled warnings are silent.

There is also a hierarchy for indicating deprecations and future
changes: for these, derive from :exc:`DeprecatedWarning`,
:exc:`MandatoryDeprecatedWarning`, :exc:`FutureDeprecatedWarning`
or :exc:`FutureReservedVariableWarning`.

Whether or not to display warnings, beyond those that are on by
default, is controlled through the command line (``--warn``) or
through ``SetOption('warn')``.  The names used there use a different
naming style than the warning class names. :func:`process_warn_strings`
converts the names before enabling/disabling.

The behavior of issuing only a message (for "enabled" warnings) can
be toggled to raising an exception instead by calling the
:func:`warningAsException` function.

For new/removed warnings, the manpage needs to be kept in sync.
Any warning class defined here is accepted, but we don't want to make
people have to dig around to find the names.  Warnings do not have to
be defined in this file, though it is preferred: those defined elsewhere
cannot use the enable/disable functionality unless they monkeypatch the
warning into this module's namespace.

You issue a warning, either in SCons code or in a build project's
SConscripts, by calling the :func:`warn` function defined in this module.
Raising directly with an instance of a warning class bypasses the
framework and it will behave like an ordinary exception.
"""

from __future__ import annotations

import sys
from typing import Callable, Sequence

import SCons.Errors

# _enabled is a list of 2-tuples with a warning class object and a
# boolean (True if that warning is enabled). Initialized in SCons/Main.py.
_enabled = []

# If False, just emit the msg for an enabled warning; else raise exception
_warningAsException: bool = False

# Function to emit the warning. Initialized by SCons/Main.py for regular use;
# the unit test will set to a capturing version for testing.
_warningOut: Callable | None = None


class SConsWarning(SCons.Errors.UserError):
    """Base class for all SCons warnings."""

class WarningOnByDefault(SConsWarning):
    """Base class for SCons warnings that are enabled by default."""

SConsWarningOnByDefault = WarningOnByDefault  # transition to new name


class LinkWarning(WarningOnByDefault):
    """Base class for linker warnings."""

# NOTE:  If you add a new warning class here, add it to the man page, too!

# General warnings

class CacheVersionWarning(WarningOnByDefault):
    """The derived-file cache directory has an out of date config."""

class CacheWriteErrorWarning(SConsWarning):
    """Problems writing a derived file to the cache."""

class CacheCleanupErrorWarning(SConsWarning):
    """Problems removing retrieved target prior to rebuilding."""

class CorruptSConsignWarning(WarningOnByDefault):
    """Problems decoding the contents of the sconsign database."""

class DependencyWarning(SConsWarning):
    """A scanner identified a dependency but did not add it."""

class DevelopmentVersionWarning(WarningOnByDefault):
    """Use of a deprecated feature."""

class DuplicateEnvironmentWarning(WarningOnByDefault):
    """A target appears in more than one consenv with identical actions.

    A duplicate target with different rules cannot be built;
    with the same rule it can, but this could indicate a problem in
    the build configuration.
    """

class FortranCxxMixWarning(LinkWarning):
    """Fortran and C++ objects appear together in a link line.

    Some compilers support this, others do not.
    """

class FutureReservedVariableWarning(WarningOnByDefault):
    """Setting a variable marked to become reserved in a future release."""

class MisleadingKeywordsWarning(WarningOnByDefault):
    """Use of possibly misspelled kwargs in Builder calls."""

class MissingSConscriptWarning(WarningOnByDefault):
    """The script specified in an SConscript() call was not found.

    TODO: this is now an error, so no need for a warning. Left in for
    a while in case anyone is using, remove eventually.

    Manpage entry removed in 4.6.0.
    """

class NoObjectCountWarning(WarningOnByDefault):
    """Object counting (debug mode) could not be enabled."""

class NoParallelSupportWarning(WarningOnByDefault):
    """Fell back to single-threaded build, as no thread support found."""

class ReservedVariableWarning(WarningOnByDefault):
    """Attempt to set reserved construction variable names."""

class StackSizeWarning(WarningOnByDefault):
    """Requested thread stack size could not be set."""

class TargetNotBuiltWarning(SConsWarning): # TODO: should go to OnByDefault
    """A target build indicated success but the file is not found."""

class VisualCMissingWarning(WarningOnByDefault):
    """Requested MSVC version not found and policy is to not fail."""

class VisualVersionMismatch(WarningOnByDefault):
    """``MSVC_VERSION`` and ``MSVS_VERSION`` do not match.

    Note ``MSVS_VERSION`` is deprecated, use ``MSVC_VERSION``.
    """

class VisualStudioMissingWarning(SConsWarning):  # TODO: unused
    pass


# Deprecation warnings

class FutureDeprecatedWarning(SConsWarning):
    """Base class for features that will become deprecated in a future release."""

class DeprecatedWarning(SConsWarning):
    """Base class for deprecated features, will be removed in future."""

class MandatoryDeprecatedWarning(DeprecatedWarning):
    """Base class for deprecated features where warning cannot be disabled."""


# Special case; base always stays DeprecatedWarning
class PythonVersionWarning(DeprecatedWarning):
    """SCons was run with a deprecated Python version."""

class DeprecatedOptionsWarning(MandatoryDeprecatedWarning):
    """Options that are deprecated."""

class DeprecatedDebugOptionsWarning(MandatoryDeprecatedWarning):
    """Option-arguments to --debug that are deprecated."""

class ToolQtDeprecatedWarning(DeprecatedWarning):  # TODO: unused
    pass


def suppressWarningClass(clazz):
    """Suppresses all warnings of type *clazz* or derived from *clazz*."""
    _enabled.insert(0, (clazz, False))

def enableWarningClass(clazz):
    """Enables all warnings of type *clazz* or derived from *clazz*."""
    _enabled.insert(0, (clazz, True))

def warningAsException(flag = True):
    """Sets global :data:`_warningAsExeption` flag.

    If true, any enabled warning will cause an exception to be raised.

    Args:
        flag: new value for warnings-as-exceptions.

    Returns:
        The previous value.
    """
    global _warningAsException  # pylint: disable=global-statement
    old = _warningAsException
    _warningAsException = flag
    return old

def warn(clazz, *args):
    """Issue a warning, accounting for SCons rules.

    Check if warnings for this class are enabled.  If warnings are treated
    as exceptions, raise exception.  Use the global warning emitter
    :data:`_warningOut`, which allows selecting different ways of
    presenting a traceback (see Script/Main.py).
    """
    warning = clazz(args)
    for cls, flag in _enabled:
        if isinstance(warning, cls):
            if flag:
                if _warningAsException:
                    raise warning

                if _warningOut is not None:
                    _warningOut(warning)
            break

def process_warn_strings(arguments):
    """Process requests to enable/disable warnings.

    The requests come from the option-argument string passed to the
    ``--warn`` command line option or as the value passed to the
    ``SetOption`` function with a first argument of ``warn``;


    The arguments are expected to be as documented in the SCons manual
    page for the ``--warn`` option, in the style ``some-type``,
    which is converted here to a camel-case name like ``SomeTypeWarning``,
    to try to match the warning classes defined here, which are then
    passed to :func:`enableWarningClass` or :func:`suppressWarningClass`.

    For example, a string``"deprecated"`` enables the
    :exc:`DeprecatedWarning` class, while a string``"no-dependency"``
    disables the :exc:`DependencyWarning` class.

    As a special case, the string ``"all"`` disables all warnings and
    a the string ``"no-all"`` disables all warnings.
    """
    def _classmunge(s):
        """Convert a warning argument to SConsCase.

        The result is CamelCase, except "Scons" is changed to "SCons"
        """
        s = s.replace("-", " ").title().replace(" ", "")
        return s.replace("Scons", "SCons")

    for arg in arguments:
        enable = True
        if arg.startswith("no-") and arg not in (
            "no-object-count",
            "no-parallel-support",
        ):
            enable = False
            arg = arg[len("no-") :]
        if arg == 'all':
            class_name = "SConsWarning"
        else:
            class_name = _classmunge(arg) + 'Warning'
        try:
            clazz = globals()[class_name]
        except KeyError:
            sys.stderr.write(f"No warning type: {arg!r}\n")
        else:
            if enable:
                enableWarningClass(clazz)
            elif issubclass(clazz, MandatoryDeprecatedWarning):
                sys.stderr.write(f"Can not disable mandataory warning: {arg!r}\n")
            else:
                suppressWarningClass(clazz)

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


# pylint: disable=missing-module-docstring,protected-access,used-before-assignment


# Nuitka will optimize this away, but VS code will warn about them otherwise.
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    compiled_function_tables = {}

try:
    import cloudpickle
except ImportError:
    pass
else:

    def _create_compiled_function(module_name, func_values):
        if module_name not in compiled_function_tables:
            __import__(module_name)

        # This gets the "_create_compiled_function" of the module and calls it.
        return compiled_function_tables[module_name][1](*func_values)

    orig_dynamic_function_reduce = (
        cloudpickle.cloudpickle.Pickler._dynamic_function_reduce
    )

    def _dynamic_function_reduce(self, func):
        if type(func).__name__ != "compiled_function":
            return orig_dynamic_function_reduce(self, func)

        try:
            module_name = func.__module__
        except AttributeError:
            return orig_dynamic_function_reduce(self, func)
        else:
            if module_name not in compiled_function_tables:
                return orig_dynamic_function_reduce(self, func)

            return (
                _create_compiled_function,
                (
                    module_name,
                    # This gets the "_reduce_compiled_function" of the module and calls it.
                    compiled_function_tables[module_name][0](func),
                ),
            )

    cloudpickle.cloudpickle.Pickler._dynamic_function_reduce = _dynamic_function_reduce

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Core formatting utility for Nuitka."""

import re
import subprocess

from nuitka.utils.FileOperations import (
    getFileContentByLine,
    getFileContents,
    putTextFileContents,
)
from nuitka.utils.PrivatePipSpace import (
    getClangFormatBinaryPath,
    withPrivatePipSitePackagesPathAdded,
)


def cleanupWindowsNewlines(filename, effective_filename):
    """Remove Windows new-lines from a file."""

    with open(filename, "rb") as f:
        source_code = f.read()

    updated_code = source_code.replace(b"\r\n", b"\n")
    updated_code = updated_code.replace(b"\n\r", b"\n")

    # Smuggle consistency replacement in here.
    if (
        "AutoFormat.py" not in effective_filename
        and "Formatting.py" not in effective_filename
    ):
        updated_code = updated_code.replace(b'.decode("utf-8")', b'.decode("utf8")')
        updated_code = updated_code.replace(b'.encode("utf-8")', b'.encode("utf8")')
        updated_code = updated_code.replace(b"# spellchecker", b"# spell-checker")

        def replacer(match):
            return b"PYTHON_VERSION %s %s" % (match.group(1), match.group(2).lower())

        updated_code = re.sub(
            b"PYTHON_VERSION\\s+([=<>]+)\\s+(0x3[A-F])", replacer, updated_code
        )

    if updated_code != source_code:
        with open(filename, "wb") as out_file:
            out_file.write(updated_code)


def cleanupTrailingWhitespace(filename):
    """Remove trailing white spaces from a file."""

    source_lines = list(getFileContentByLine(filename, encoding="utf8"))

    clean_lines = [line.rstrip().replace("\t", "    ") for line in source_lines]

    while clean_lines and clean_lines[-1] == "":
        del clean_lines[-1]

    if clean_lines != source_lines or (clean_lines and clean_lines[-1] != ""):
        putTextFileContents(filename, contents=clean_lines, encoding="utf8")


_warned_clang_format = False
_clang_format_path = False


def _getClangFormatPath(logger, assume_yes_for_downloads, reject_message):
    """Get the path to the clang-format executable.

    Args:
        logger: logger to use
        assume_yes_for_downloads: bool
    """
    # pylint: disable=global-statement
    global _warned_clang_format, _clang_format_path

    if _warned_clang_format:
        return None

    if _clang_format_path is not False:
        return _clang_format_path

    _clang_format_path = getClangFormatBinaryPath(
        logger=logger,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )

    if _clang_format_path is None and not _warned_clang_format:
        if logger is not None:
            logger.warning("Need to accept clang-format download to format C files.")
        _warned_clang_format = True

    return _clang_format_path


def _cleanupClangFormat(logger, filename, assume_yes_for_downloads, reject_message):
    """Call clang-format on a given filename to format C code.

    Args:
        filename: path to the file
        logger: logger to use
        assume_yes_for_downloads: bool
    """

    clang_format_path = _getClangFormatPath(
        logger=logger,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )

    if clang_format_path:
        with withPrivatePipSitePackagesPathAdded(logger=logger):
            subprocess.call(
                [
                    clang_format_path,
                    "-i",
                    "-style={BasedOnStyle: llvm, IndentWidth: 4, ColumnLimit: 120}",
                    filename,
                ]
            )


def formatC(
    logger,
    filename,
    effective_filename,
    check_only,
    assume_yes_for_downloads,
    reject_message,
):
    """Format C/C++ source code.

    Args:
        filename: path to the file
        effective_filename: filename to use for errors
        logger: logger to use
        check_only: bool - if only checking is to be done
        assume_yes_for_downloads: bool - if downloads are allowed
    """
    if check_only:
        try:
            getFileContents(filename, encoding="ascii")
        except UnicodeDecodeError:
            if logger:
                logger.warning(
                    "All C files must be pure ASCII, need to convert it manually."
                )
            return True

    cleanupWindowsNewlines(filename, effective_filename)
    _cleanupClangFormat(
        logger=logger,
        filename=filename,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )
    cleanupWindowsNewlines(filename, effective_filename)

    return False


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
